"use strict";

const EXPORTED_SYMBOLS = ["barApplication"];

const { classes: Cc, interfaces: Ci, utils: Cu } = Components;
var sysutils,
    fileutils,
    strutils,
    xmlutils;

const barApplication = {
    init: function BarApp_init(core) {
        this._barCore = core;
        let (appCoreLib = core.Lib) {
            sysutils = appCoreLib.sysutils;
            fileutils = appCoreLib.fileutils;
            strutils = appCoreLib.strutils;
            xmlutils = appCoreLib.xmlutils;
        };
        
        this._logger = this._barCore.Lib.Log4Moz.repository.getLogger(core.appName + ".App");
        this._dirs._barApp = this;
        this._wndControllerName = this.name + "OverlayController";
        
        this._init();
        this._parts.installer.saveBuildDataToPreferences();
        
        try {
            this.widgetLibrary.persist(false, true);
        }
        catch (e) {
            this._logger.error("Failed writing plugins list. " + this._barCore.Lib.strutils.formatError(e));
        }
        
        let cleanupFunc = this._barCore.Lib.sysutils.bindFunction(
            function _cleanupPreferences_timed() {
                try {
                    this._logger.debug("Preferences cleanup started");
                    this._cleanupPreferences();
                } catch (e) {
                    this._logger.error("Failed cleaning preferences. " + this._barCore.Lib.strutils.formatError(e));
                }
            }, this);
        new this._barCore.Lib.sysutils.Timer(cleanupFunc, 2 * 60 * 1000);
    },
    
    finalize: function BarApp_finalize() {
        let doFinalCleanup = this._parts.installer.isAddonUninstalling;
        let addonId = this._parts.installer.addonId;
        this._finalizeParts(doFinalCleanup);
        
        if (doFinalCleanup)
            this._finalCleanup(addonId);
        
        this._logger = null;
        this._barCore = null;
    },
    get defaultPresetURL() {
        return this._barCore.Lib.Preferences.get(this.name + this._consts.PREF_DEFAULT_PRESET_URL, null);
    },
    
    get defaultPreset() {
        return this._defaultPreset;
    },
    
    set defaultPreset(newPreset) {
        if (!(newPreset instanceof this.BarPlatform.Preset))
            throw new this._barCore.Lib.CustomErrors.EArgType("newPreset", "Preset", newPreset);
        this._defaultPreset = newPreset;
    },
    
    get name() {
        return this._barCore.appName;
    },
    get preferencesBranch() {
        let appPrefsBranch = "extensions." + this._parts.installer.addonId + ".";
        delete this.preferencesBranch;
        this.__defineGetter__("preferencesBranch", function() appPrefsBranch);
        return appPrefsBranch;
    },
    
    get preferences() {
        let appPrefs = new this._barCore.Lib.Preferences(this.preferencesBranch);
        delete this.preferences;
        this.__defineGetter__("preferences", function() appPrefs);
        return appPrefs;
    },
    
    _delayMultiplier: 0,
    _lastGeneratedDelay: 0,
    generateDelay: function BarApp_generateDelay() {
        if (!this._delayMultiplier) {
            this._delayMultiplier = this._barCore.Lib.Preferences.get(this.name + ".debug.delayMultiplier", 60);
        }
        this._lastGeneratedDelay += this._delayMultiplier;
        return this._lastGeneratedDelay;
    },
    
    getNewWidgetInstanceID: function barApp_getNewWidgetInstanceID() {
        return "" + Date.now() + this._newWID++;
    },
    
    genWindowEngineID: function BarApp_genWindowEngineID() {
        return this._newWEID++;
    },
    
    getLogger: function BarApp_getLogger(name) {
        return this._barCore.Lib.Log4Moz.repository.getLogger(this.name + "." + name);
    },
    get localeString() {
        if (!this._localeString) {
            let xulChromeReg = Cc["@mozilla.org/chrome/chrome-registry;1"].getService(Ci.nsIXULChromeRegistry);
            try {
                this._localeString = xulChromeReg.getSelectedLocale(this.name);
            } catch (ex) {
                this._localeString = xulChromeReg.getSelectedLocale("global");
            }
        }
        
        return this._localeString || "ru";
    },
    
    get locale() {
        if (!this._locale)
            this._locale = this._barCore.Lib.misc.parseLocale(this.localeString);
        
        return this._locale;
    },
    
    get soundsEnabled() {
        return this.preferences.get("sounds.enabled", true);
    },
    
    get core() {
        return this._barCore;
    },
    
    get directories() {
        return this._dirs;
    },
    
    get partsURL() {
        return "resource://" + this.name + "-app/parts/";
    },
    
    get packageManager() {
        return this._parts.pacMan;
    },
    
    get widgetLibrary() {
        return this._parts.widgetLibrary;
    },
    
    get overlayProvider() {
        return this._parts.overlayProvider;
    },
    
    get addonFS() {
        return this._parts.addonFS;
    },
    
    get BarPlatform() {
        return this._parts.BarPlatform;
    },
    
    get XB() {
        return this._parts.XB;
    },
    
    get UIBuilder() {
        return this._parts.UI.Builder;
    },
    
    get WindowEngine() {
        return this._parts.WindowEngine;
    },
    
    get StringBundle() {
        return this._parts.StringBundlePart.StringBundle;
    },
    
    get NativeComponents() {
        return this._parts.NativeComponents;
    },
    
    get statLogger() {
        return this._parts.statLogger;
    },
    
    get installer() {
        return this._parts.installer;
    },
    
    get updater() {
        return this._parts.updater;
    },
    
    get branding() {
        return this._parts.branding;
    },
    
    get packagesBlackList() {
        if (this._blackList)
            return this._blackList;
        
        let blFile = this._blackListFile;
        
        if (!blFile.exists()) {
            return this.packagesBlackList = this._getInternalBlackList();
        }
        
        try {
            let blackListJson = this._barCore.Lib.fileutils.jsonFromFile(blFile);
            return this._blackList = new this.BarPlatform.BlackList.fromJSON(blackListJson);
        } catch(e) {
            this._logger.error("Could not read packages black list \n" + this._barCore.Lib.strutils.formatError(e));
            return this.packagesBlackList = this._getInternalBlackList();
        }
    },
    
    set packagesBlackList(newBlackList) {
        this._barCore.Lib.fileutils.jsonToFile(newBlackList, this._blackListFile);
        this._blackList = newBlackList;
    },
    
    restartComponents: function BarApp_restartComponents(packageID) {
        this.switchWidgets(packageID, false);
        let pluginsState = this.widgetLibrary.getCurrentPluginsState();
        this.widgetLibrary.flushPlugins(packageID);
        
        this.packageManager.reloadPackage(packageID);
        
        this.widgetLibrary.setPluginsState(pluginsState);
        this.switchWidgets(packageID, true);
    },
    
    switchWidgets: function BarApp_switchWidgets(packageID, on) {
        this.forEachWindow(
            function(controller) {
                controller.switchWidgets(packageID, on);
            });
    },
    
    forEachWindow: function BarApp_forEachWindow(func, contextObj, handleExceptions) {
        if (typeof func !== "function")
            throw new this._barCore.Lib.CustomErrors.EArgType("func", "Function", func);
        
        let browserWindows = this.core.Lib.misc.getBrowserWindows();
        for (let i = browserWindows.length; i--;) {
            try {
                let controller = browserWindows[i][this._wndControllerName];
                if (controller)
                    func.call(contextObj, controller);
            }
            catch (e) {
                if (!handleExceptions)
                    throw e;
                
                let errMsg = this._barCore.Lib.strutils.
                        formatString("Could not call browser controller. Caller: '%1'. %2",
                                     [func.name, e]);
                this._logger.error(errMsg);
                this._logger.debug(e.stack);
            }
        }
    },
    
    installPreset: function BarApp_installPreset(url) {
        this._barCore.Lib.misc.openWindow({
            url: "chrome://" + this.name + "/content/custombar/dialogs/package-management/install/install.xul",
            features: "__popup__",
            name: "package-management-install",
            mode: "install",
            preset: url,
            application: this
        });
    },
    
    
    openSettingsDialog: function BarApp_openSettingsDialog(navigatorWindow ) {
        let chromePath = "chrome://" + this.name + "/content/custombar/preferences/preferences.xul",
            windowClass = this.name + ":Preferences",
            resizeable = true,
            modal = !this._barCore.Lib.Preferences.get("browser.preferences.instantApply", false),
            windowArgs = Array.slice(arguments, 1),
            focusIfOpened = true;
        
        this._openWindow.apply(this, [navigatorWindow, chromePath, windowClass, focusIfOpened, resizeable, modal, windowArgs]);
    },
    
    openAboutDialog: function BarApp_openAboutDialog() {
        let chromePath = "chrome://" + this.name + "/content/custombar/dialogs/about/about.xul",
            windowClass = this.name + ":AboutDialog",
            focusIfOpened = true;
        
        this._openWindow(null, chromePath, windowClass, focusIfOpened);
    },
    
    
    navigate: function BarApp_navigate(aNavigateData) {
        
        if (typeof aNavigateData != "object")
            throw new Error("Object required.");
        
        let url;
        let unsafe = false;
        
        if ("url" in aNavigateData) {
            url = aNavigateData.url;
        } else if ("unsafeURL" in aNavigateData) {
            url = aNavigateData.unsafeURL;
            unsafe = true;
        }
        if (!url)
            return false;
        let uri = this._barCore.Lib.misc.tryCreateFixupURI(url);
        
        if (!uri)
            throw new this._barCore.Lib.CustomErrors.EArgRange("url", "URL", url);
        
        if (unsafe && !/^(http|ftp)s?$/.test(uri.scheme))
            throw new this._barCore.Lib.CustomErrors.ESecurityViolation("application.navigate", "URL=" + url);
        
        url = uri.spec;
        
        let target = ("target" in aNavigateData) ? aNavigateData.target : null;
        
        if (!target) {
            let eventInfo = ("eventInfo" in aNavigateData) ? aNavigateData.eventInfo : null;
            
            if (eventInfo) {
                if (eventInfo instanceof Ci.nsIDOMEvent) {
                    if (eventInfo.ctrlKey || eventInfo.metaKey || eventInfo.button == 1)
                        target = "new tab";
                    else if (eventInfo.shiftKey)
                        target = "new window";
                } else {
                    if (eventInfo.keys.ctrl || eventInfo.keys.meta || eventInfo.mouse.button == 1)
                        target = "new tab";
                    else if (eventInfo.keys.shift)
                        target = "new window";
                }
            }
        }
        
        let postData = ("postData" in aNavigateData) ? aNavigateData.postData : null;
        let referrer = ("referrer" in aNavigateData) ? aNavigateData.referrer : null;
        let sourceWindow = ("sourceWindow" in aNavigateData) ? aNavigateData.sourceWindow : null;
        if (!sourceWindow && target != "new popup") {
            sourceWindow = this.core.Lib.misc.getTopBrowserWindow();
            
            if (!sourceWindow) {
                let sa = Cc["@mozilla.org/supports-array;1"].createInstance(Ci.nsISupportsArray);
                
                let wuri = Cc["@mozilla.org/supports-string;1"].createInstance(Ci.nsISupportsString);
                wuri.data = url;
                
                let allowThirdPartyFixupSupports = Cc["@mozilla.org/supports-PRBool;1"].createInstance(Ci.nsISupportsPRBool);
                allowThirdPartyFixupSupports.data = false;
                
                sa.AppendElement(wuri);
                sa.AppendElement(null);
                sa.AppendElement(referrer);
                sa.AppendElement(postData);
                sa.AppendElement(allowThirdPartyFixupSupports);
                
                let windowWatcher = Cc["@mozilla.org/embedcomp/window-watcher;1"].getService(Ci.nsIWindowWatcher);
                windowWatcher.openWindow(null, "chrome://browser/content/browser.xul", null, "chrome,dialog=no,all", sa);
                return true;
            }
        }
        if (!target)
            target = "none";
        
        if (target != "new popup") {
            sourceWindow.Ya.loadURI(url, target, null, postData);
            return;
        }
        
        switch (target) {
            case "new tab":
                sourceWindow.gBrowser.loadOneTab(url, referrer, null, postData, false);
                break;
            
            case "new window":
                sourceWindow.openNewWindowWith(url, null, postData, false, referrer);
                break;
            
            case "new popup": {
                let windowProperties = ("windowProperties" in aNavigateData) ? aNavigateData.windowProperties : {};
                
                let title = ("title" in windowProperties) ? windowProperties.title : null;
                let wndWidth = Math.max(parseInt(windowProperties.width, 10) || 300, 50);
                let wndHeight = Math.max(parseInt(windowProperties.height, 10) || 300, 50);
                
                let winFeatures = "chrome,all,dialog=no,resizable,centerscreen,width=" + wndWidth + ",height=" + wndHeight;
                
                let args = {
                    url: url,
                    title: title,
                    postData: postData,
                    referrer: referrer
                };
                args.wrappedJSObject = args;
                
                let popupChromeURL = "chrome://" + this.name + "/content/custombar/dialogs/popup_browser/popup_browser.xul";
                if (sourceWindow) {
                    sourceWindow.openDialog(popupChromeURL, null, winFeatures, args);
                } else {
                    let windowWatcher = Cc["@mozilla.org/embedcomp/window-watcher;1"].getService(Ci.nsIWindowWatcher);
                    windowWatcher.openWindow(null, popupChromeURL, null, winFeatures, args);
                }
                
                break;
            }
            default:
                sourceWindow.gBrowser.loadURI(url, referrer, postData, false);
                break;
        }
        
        return true;
    },
    
    
    selectBestPackage: function BarApp_selectBestPackage(manifest ) {
        if (!(manifest instanceof this.BarPlatform.PackageManifest))
            throw new this._barCore.Lib.CustomErrors.EArgType("manifest", "PackageManifest", manifest);
        
        const sysutils = this._barCore.Lib.sysutils;
        let platformInfo = sysutils.platformInfo,
            coreVersion = this._barCore.version,
            bestPackageInfo = null,
            bestVersion = "0",
            bestPackageInfo2 = null,
            bestVersion2 = "0";
        
        let comparator = sysutils.versionComparator;
        
        for each (let packageInfo in manifest.packagesInfo) {
            if (packageInfo.browser && (packageInfo.browser != platformInfo.browser.name)) {
                this._logger.warn("Package [" + packageInfo.id + "] for browser [" + packageInfo.browser + "] is not usable on [" + platformInfo.browser.name + "]");
                continue;
            }
            
            if (packageInfo.os && (packageInfo.os != platformInfo.os.name)) {
                this._logger.warn("Package [" + packageInfo.id + "] for os [" + packageInfo.os + "] is not usable on [" + platformInfo.os.name + "]");
                continue;
            }
            
            if (packageInfo.architecture && (packageInfo.architecture != platformInfo.browser.simpleArchitecture)) {
                this._logger.warn("Package [" + packageInfo.id + "] for architecture [" + packageInfo.architecture + "] is not usable on [" + platformInfo.browser.simpleArchitecture + "]");
                continue;
            }
            
            let banReason = this.packagesBlackList.banReasonForPackage(packageInfo);
            if (banReason) {
                let banMessage = this.packagesBlackList.messageForReasonByLocale(banReason, this.localeString);
                this._logger.warn("Package [" + packageInfo.id + "] with version [" + packageInfo.version + "] is banned (" + banMessage + ")");
                continue;
            }
            
            if (packageInfo.platformMin <= coreVersion) {
                let newerPackage = (comparator.compare(packageInfo.version, bestVersion) > 0),
                    newerPlatform = bestPackageInfo && (packageInfo.platformMin > bestPackageInfo.platformMin);
                if (newerPackage || newerPlatform) {
                    bestPackageInfo = packageInfo;
                    bestVersion = packageInfo.version;
                }
            }
            else {
                if (comparator.compare(packageInfo.version, bestVersion2) > 0) {
                    bestPackageInfo2 = packageInfo;
                    bestVersion2 = packageInfo.version;
                }
            }
        }
        return [bestPackageInfo, bestPackageInfo2];
    },
    
    isYandexHost: function BarApp_isYandexHost(hostName ) {
        return this._yandexHostsPattern.test(hostName);
    },
    isPrivilegedPackageURL: function BarApp_isPrivilegedPackageURL(url ) {
        let ioService = Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService);
        return this.isPrivilegedPackageURI(ioService.newURI(url, null, null));
    },
    
    isPrivilegedPackageURI: function BarApp_isPrivilegedPackageURI(uri ) {
        if (!(uri instanceof Ci.nsIURI))
            throw new this._barCore.Lib.CustomErrors.EArgType("uri", "nsIURI", uri);
        
        if ((uri.scheme != "http") && (uri.scheme != "https")) {
            return false;
        }
        
        let trust = ((uri.host == this._consts.WIDGET_LIBRARY_HOST) && (uri.path.indexOf("/packages/approved/") == 0)) ||
                    ((uri.host in this._consts.BAR_HOSTS) && (uri.path.indexOf("/packages/") == 0)) ||
                    (uri.host == this._consts.DOWNLOAD_HOST_NAME) ||
                    (uri.host == this._consts.BARQA_HOST_NAME && uri.path.indexOf("/notapp/") == -1);
        
        return trust;
    },
    isTrustedPackageURL: function BarApp_isTrustedPackageURL(url ) {
        let ioService = Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService);
        return this.isTrustedPackageURI(ioService.newURI(url, null, null));
    },
    
    isTrustedPackageURI: function BarApp_isTrustedPackageURI(uri ) {
        return (uri.host == this._consts.WIDGET_LIBRARY_HOST) && (uri.path.indexOf("/packages/") === 0) || this.isPrivilegedPackageURI(uri);
    },
    isTrustedPresetUri: function BarApp_isTrustedPresetUri(uri ) {
        if (!(uri instanceof Ci.nsIURI))
            throw new this._barCore.Lib.CustomErrors.EArgType("uri", "nsIURI", uri);
        
        if ((uri.scheme != "http") && (uri.scheme != "https"))
            return false;
        
        return ((uri.host == this._consts.WIDGET_LIBRARY_HOST) && (uri.path.indexOf("/presets/approved/") == 0));
    },
    
    onNewBrowserReady: function BarApp_onNewBrowserReady(controller) {
        
        var isFirstWindow = (++this._navigatorID == 1);
        
        if (isFirstWindow) {
            try {
                this._parts.preinstaller.activateIncoming(controller);
            }
            catch (e) {
                this._logger.error("Could not place preinstalled widgets on toolbar. " + this._barCore.Lib.strutils.formatError(e));
            }
            
            try {
                this._parts.daylystat.onFisrtNavigatorReady();
            }
            catch (e) {
                this._logger.error("Could not notify daylystat. " + this._barCore.Lib.strutils.formatError(e));
                this._logger.debug(e.stack);
            }
            
            try {
                let autoPresetURL = this._autoPresetURL;
                if (autoPresetURL) {
                    this._removeAutoPresetCookie();
                    this.installPreset(autoPresetURL);
                }
            }
            catch (e) {
                this._logger.error("Could not check autoinstalled preset. " + this._barCore.Lib.strutils.formatError(e));
                this._logger.debug(e.stack);
            }
        }
    },
    _consts: {
        PREF_DEFAULT_PRESET_URL: ".default.preset.url",
        DEF_PRESET_FILE_NAME: "default.xml",
        DOWNLOAD_HOST_NAME: "download.yandex.ru",
        BARQA_HOST_NAME: "bar.qa.yandex.net",
        BAR_HOSTS: {
            "bar.yandex.ru": 0,
            "toolbar.yandex.ru": 0
        },
        WIDGET_LIBRARY_HOST: "bar-widgets.yandex.ru",
        AUTOPRESET_COOKIE_DOMAIN: "bar.yandex.ru",
        AUTOPRESET_COOKIE_NAME: "bar-install-preset"
    },
    _barCore: null,
    _logger: null,
    _defaultPreset: null,
    _usingInternalPreset: false,
    _introducedPlugins: [],
    _blackList: null,
    _overlayProvider: null,
    _navigatorID: 0,
    _newWID: 0,
    _newWEID: 0,
    _localeString: null,
    _wndControllerName: undefined,
    _yandexHostsPattern: /(^|\.)yandex\.(ru|ua|by|kz|net|com)$/i,
    _dirs: {
        get appRootDir() {
            if (!this._appRoot)
                this._appRoot = this._barApp.core.rootDir;
            let dirFile = this._appRoot.clone();
            this._forceDir(dirFile);
            return dirFile;
        },
        
        get packagesDir() {
            let packagesDir = this.appRootDir;
            packagesDir.append("packages");
            this._forceDir(packagesDir);
            return packagesDir;
        },
        
        get parsedCompsDir() {
            let parsedDir = this.appRootDir;
            parsedDir.append("parsed_comps");
            this._forceDir(parsedDir);
            return parsedDir;
        },
        
        get presetsDir() {
            let presetsDir = this.appRootDir;
            presetsDir.append("presets");
            this._forceDir(presetsDir);
            return presetsDir;
        },
        
        get nativeStorageDir() {
            let storageDir = this.appRootDir;
            storageDir.append("native_storage");
            this._forceDir(storageDir);
            return storageDir;
        },
        
        get vendorDir() {
            let vendorDir = this.appRootDir;
            vendorDir.append("vendor");
            this._forceDir(vendorDir);
            return vendorDir;
        },
        
        makePackageDirName: function BarApp_makePackageDirName() {
            return this._uuidGen.generateUUID().toString();
        },
        
        _barApp: null,
        _uuidGen: Cc["@mozilla.org/uuid-generator;1"].getService(Ci.nsIUUIDGenerator),
        
        _forceDir: function BarAppDirs_forceDir(dirFile, perm) {
            this._barApp.core.Lib.fileutils.forceDirectories(dirFile, perm);
        }
    },
    _parts: {},
    _partsNames: {
        installer:          "installer.js",
        addonFS:            "addonfs.js",
        StringBundlePart:   "strbundle.js",
        statLogger:         "statlogger.js",
        daylystat:          "daylystat.js",
        BarPlatform:        "platform.js",
        XB:                 "xb.js",
        UI:                 "uibuilder.js",
        NativeComponents:   "native_comps.js",
        branding:           "partnerpack.js",
        pacMan:             "pacman.js",
        widgetLibrary:      "widgetlib.js",
        overlayProvider:    "overlay_prov.js",
        WindowEngine:       "wndengine.js",
        preinstaller:       "preinstaller.js",
        updater:            "update.js"
    },
    
    _PREV_BUILTIN_WIDS: [
        "http://bar.yandex.ru/packages/yandexbar#logo",
        "http://bar.yandex.ru/packages/yandexbar#search",
        "http://bar.yandex.ru/packages/yandexbar#cy",
        "http://bar.yandex.ru/packages/yandexbar#spam",
        "http://bar.yandex.ru/packages/yandexbar#opinions",
        "http://bar.yandex.ru/packages/yandexbar#spellchecker",
        "http://bar.yandex.ru/packages/yandexbar#translator",
        "http://bar.yandex.ru/packages/yandexbar#pagetranslator",
        "http://bar.yandex.ru/packages/yandexbar#login",
        "http://bar.yandex.ru/packages/yandexbar#mail",
        "http://bar.yandex.ru/packages/yandexbar#fotki",
        "http://bar.yandex.ru/packages/yandexbar#yaru",
        "http://bar.yandex.ru/packages/yandexbar#moikrug",
        "http://bar.yandex.ru/packages/yandexbar#lenta",
        "http://bar.yandex.ru/packages/yandexbar#money",
        "http://bar.yandex.ru/packages/yandexbar#zakladki",
        "http://bar.yandex.ru/packages/yandexbar#widget-news",
        "http://bar.yandex.ru/packages/yandexbar#settings",
        "http://bar.yandex.ru/packages/yandexbar#partner"
    ],
    
    get _blackListFile() {
        let blFile = this.directories.appRootDir;
        blFile.append("black-list.json");
        return blFile;
    },
    
    _getInternalBlackList: function BarApp__getInternalBlackList() {
        try {
            let blStream = this.addonFS.getStreamFromPath("$custombar/packages/black-list.xml");
            let doc = this._barCore.Lib.fileutils.xmlDocFromStream(blStream);
            return new this.BarPlatform.BlackList.fromDoc(doc);
        }
        catch (e) {
            this._logger.error("Could not get internal packages black list.\n" + this._barCore.Lib.strutils.formatError(e));
            return new this.BarPlatform.BlackList.fromNull;
        }
    },
    
    get _autoPresetURL() {
        
        const ioService = Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService);
        let autoPresetCookieURI = ioService.newURI("http://" + this._consts.AUTOPRESET_COOKIE_DOMAIN + "/", null, null);
        const cookieService = Cc["@mozilla.org/cookieService;1"].getService().QueryInterface(Ci.nsICookieService);
        let cookieString = cookieService.getCookieString(autoPresetCookieURI, null);
        
        let presetURL = "";
        
        if (cookieString) {
            let cookies = cookieString.split(";");
            for each (let cookie in cookies) {
                let separatorPos = cookie.indexOf("=");
                if (separatorPos == -1)
                    continue;
                
                let cookieName = cookie.substring(0, separatorPos).trim();
                if (cookieName == this._consts.AUTOPRESET_COOKIE_NAME) {
                    presetURL = unescape( cookie.substring(separatorPos + 1 ) );
                    break;
                }
            }
        }
        
        this.__defineGetter__("_autoPresetURL", function _autoPresetURL() presetURL);
        return this._autoPresetURL;
    },
    _removeAutoPresetCookie: function BarApp__removeAutoPresetCookie() {
        const cookieManager = Cc["@mozilla.org/cookiemanager;1"].getService(Ci.nsICookieManager);
        cookieManager.remove("." + this._consts.AUTOPRESET_COOKIE_DOMAIN, this._consts.AUTOPRESET_COOKIE_NAME, "/", false);
        
        this.__defineGetter__("_autoPresetURL", function _autoPresetURL() "");
    },
    
    
    _finalCleanup: function BarApp__finalCleanup(aAddonId) {
        this._logger.debug("Cleanup...");
        
        let Preferences = this._barCore.Lib.Preferences;
        
        let prefBranches = [this._barCore.xbWidgetsPrefsPath,
                            this._barCore.nativesPrefsPath,
                            this._barCore.staticPrefsPath,
                            this.name + ".versions." ];
        if (aAddonId)
            prefBranches.push("extensions." + aAddonId + ".");
        
        for each (let prefBranch in prefBranches) {
            try {
                Preferences.resetBranch(prefBranch);
            } catch (e) {
                this._logger.error("Final cleanup: can't reset branch '" + prefBranch + "'. " +
                                   this._barCore.Lib.strutils.formatError(e));
            }
        }
        
        let prefs = [ this.name + this._consts.PREF_DEFAULT_PRESET_URL ];
        for each (let pref in prefs) {
            try {
                Preferences.reset(pref);
            } catch (e) {
                this._logger.error("Final cleanup: can't reset pref '" + pref + "'. " +
                                   this._barCore.Lib.strutils.formatError(e));
            }
        }
        this._logger.debug("Removing all files");
        this._barCore.logging = false;
        this._barCore.Lib.fileutils.removeFileSafe(this.directories.appRootDir);
    },
    
    _init: function BarApp__init() {
        this._logger.config("Initializing...");
        let startTime = Date.now();
        try {
            this._loadParts();
        }
        catch (e) {
            this._finalizeParts();
            throw e;
        }
        this._findDefaultPreset();
        
        let installerInfo = this._parts.installer.info;
        if (installerInfo.isFreshAddonInstall) {
            try {
                this._autoInstallDefaultComponents();
            }
            catch (e) {
                this._logger.error("Failed autoinstalling default preset components. " + strutils.formatError(e));
                this._logger.debug(e.stack);
            }
        }
        else if (installerInfo.addonVersionChanged && !this._usingInternalPreset) {
            try {
                this._mergePresets();
            }
            catch (e) {
                this._logger.error("Failed merging default presets. " + strutils.formatError(e));
                this._logger.debug(e.stack);
            }
        }
        let staticWidgets = this.overlayProvider.staticWidgets;
        for (let widgetID in this._defaultPreset.widgetIDs) {
            if (!(widgetID in staticWidgets))
                this.widgetLibrary.registerWidgets(widgetID, true);
        }
        for (let pluginID in this._defaultPreset.pluginIDs) {
            this.widgetLibrary.registerPlugins(pluginID, true);
        }
        if (this._introducedPlugins.length > 0)
            this._logger.debug("Introducing plugins: " + this._introducedPlugins);
        for each (let pluginID in this._introducedPlugins) {
            try {
                let plugin = this.widgetLibrary.getPlugin(pluginID);
                plugin.enable();
            }
            catch (e) {
                this._logger.warn("Could not activate introduced plugin " + pluginID + ". " +
                                   strutils.formatError(e));
            }
        }
        
        this.widgetLibrary.activatePlugins();
        
        if (installerInfo.addonVersionChanged) {
            try {
                this._migrateSettings();
            }
            catch (e) {
                this._logger.error("Failed migrating from another version. " + strutils.formatError(e));
                this._logger.debug(e.stack);
            }
        }
        
        this._logger.config("Init done in " + (Date.now() - startTime) + "ms");
    },
    _migrateSettings: function BarApp__migrateSettings() {
        let installerInfo = this._parts.installer.info;
        
        if (installerInfo.addonDowngraded)
            return;
        
        this._logger.config("Migrating settings...");
        
        let Preferences = this._barCore.Lib.Preferences;
        let versionComparator = this._barCore.Lib.sysutils.versionComparator;
        if (versionComparator.compare(installerInfo.addonLastVersion, "5.2.0") < 0) {
            let textonlyPluginId = "http://bar.yandex.ru/packages/yandexbar#textonly";
            let oldTextonlyPrefBranch = "yasearch.general.ui.textonly.";
            let introducedTextonly = Preferences.get(oldTextonlyPrefBranch + "introduced", false);
            if (introducedTextonly) {
                let textonlyPrefPath = this.NativeComponents.makeStaticBranchPath(textonlyPluginId);
                Preferences.set(textonlyPrefPath + "introduced", true);
                Preferences.reset(oldTextonlyPrefBranch + "introduced");
            }
            
            let enabledTextonly = Preferences.get(oldTextonlyPrefBranch + "enabled", null);
            if (enabledTextonly !== null) {
                try {
                    this.widgetLibrary.getPlugin(textonlyPluginId).enabled = enabledTextonly;
                } catch (e) {
                    this._logger.config("Couldn't change Textonly plugin state. " + this._barCore.Lib.strutils.formatError(e));
                }
                Preferences.reset(oldTextonlyPrefBranch + "enabled");
            }
        }
    },
    
    _mergePresets: function BarApp__mergePresets() {
        this._logger.config("Merging presets...");
        const internalPresetPath = "$custombar/presets/" + this._consts.DEF_PRESET_FILE_NAME;
        let internalPreset = new this.BarPlatform.Preset(
                this._barCore.Lib.fileutils.xmlDocFromStream(
                    this.addonFS.getStreamFromPath(internalPresetPath) ) );
        
        let dirtyPreset = false;
        let samePresetURLs = (internalPreset.url == this._defaultPreset.url);
        for each (let [, pluginEntry] in Iterator(internalPreset.pluginEntries)) {
            if (this._defaultPreset.refsPlugin(pluginEntry.componentID))
                continue;
            
            if (!samePresetURLs)
                this._defaultPreset.appendEntry(pluginEntry);
            if (pluginEntry.enabled == pluginEntry.ENABLED_YES)
                this._introducedPlugins.push(pluginEntry.componentID);
            
            dirtyPreset = true;
        }
        
        if (samePresetURLs) {
            this._logger.config("Replacing existing default preset with internal version...");
            this._defaultPreset = internalPreset;
            
            let presetFileName = encodeURIComponent(this._defaultPreset.url);
            let presetFile = this.directories.presetsDir;
            presetFile.append(presetFileName);
            fileutils.removeFileSafe(presetFile);
            let fMask = parseInt("0755", 8);
            this.addonFS.copySource(internalPresetPath, this.directories.presetsDir, presetFileName, fMask);
            
            return;
        }
        if (sysutils.versionComparator.compare(this._defaultPreset.formatVersion, "2.0") < 0) {
            try {
                let oldPresetWidgetIDs = this._defaultPreset.widgetIDs;
                
                this._logger.config("Fixing old default preset...");
                let CompEntryProps = this._defaultPreset.constructor.ComponentEntry.prototype;
                for each (let [, invisWidgetID] in Iterator(this._PREV_BUILTIN_WIDS)) {
                    if (invisWidgetID in oldPresetWidgetIDs) continue;
                    let entryDescr = {
                        componentType: CompEntryProps.TYPE_WIDGET,
                        componentID: invisWidgetID,
                        enabled: CompEntryProps.ENABLED_NO
                    };
                    this._defaultPreset.appendEntry(entryDescr);
                }
                this._defaultPreset.formatVersion = "2.0";
                
                dirtyPreset = true;
            }
            catch (e) {
                this._logger.fatal("Could not fix default preset. " + this._barCore.Lib.strutils.formatError(e));
            }
        }
        if (dirtyPreset) {
            let presetFile = this.directories.presetsDir;
            presetFile.append(encodeURIComponent(this._defaultPreset.url));
            this._defaultPreset.saveToFile(presetFile);
        }
    },
    
    _autoInstallDefaultComponents: function BarApp__autoInstallDefaultComponents() {
        try {
            const partsDirPath = "resource://" + this.name + "-app/parts/";
            let autoInstNS = {};
            Cu.import(partsDirPath + "autoinst.js", autoInstNS);
            let autoInstPart = autoInstNS.autoinstaller;
            autoInstPart.init(this);
            autoInstPart.activateComponents(this.defaultPreset);
        }
        catch (e) {
            this._logger.error("Could not modify default preset. " + strutils.formatError(e));
            this._logger.debug(e.stack);
        }
        finally {
            const separatorID = "http://bar.yandex.ru/packages/yandexbar#separator";
            const searchFieldID = "http://bar.yandex.ru/packages/yandexbar#search";
            let separatedVisWidgets = 0;
            for each (let compEntry in this._defaultPreset.allEntries) {
                if (compEntry.enabled == compEntry.ENABLED_AUTO) {
                    compEntry.enabled = compEntry.ENABLED_NO;
                }
                
                let entryID = compEntry.componentID;
                let isSearchField = entryID == searchFieldID;
                if (isSearchField || (entryID == separatorID)) {
                    if (separatedVisWidgets > 0)
                        separatedVisWidgets = 0;
                    else if (!isSearchField)
                        compEntry.enabled = compEntry.ENABLED_NO;
                }
                else if (compEntry.enabled == compEntry.ENABLED_YES) {
                    separatedVisWidgets++;
                }
            }
        }
        
        let presetFile = this.directories.presetsDir;
        presetFile.append(encodeURIComponent(this._defaultPreset.url));
        this._defaultPreset.saveToFile(presetFile);
    },
    
    _loadParts: function BarApp__loadParts() {
        const Cu = Components.utils;
        const partsDirPath = "resource://" + this.name + "-app/parts/";
        
        for (let partName in this._partsNames) {
            let partFileName = this._partsNames[partName];
            let partPath = partsDirPath + partFileName;
            this._logger.debug("Loading " + partName + " part from " + partPath);
            Cu.import(partPath, this._parts);
            let part = this._parts[partName];
            if (!part)
                throw new Error("Part " + partName + " not loaded!");
            if (typeof part.init == "function")
                part.init(this);
        }
    },
    
    _finalizeParts: function BarApp__finalizeParts(doCleanup) {
        let revParts = this._barCore.Lib.misc.mapKeysToArray(this._parts).reverse();
        for each (let partName in revParts) {
            this._logger.debug("Finalizing " + partName + " part");
            let part = this._parts[partName];
            if (typeof part.finalize == "function") {
                try {
                    part.finalize(doCleanup);
                    this._parts[partName] = null;
                }
                catch (e) {
                    this._logger.error("Error finalizing part. " + this._barCore.Lib.strutils.formatError(e));
                    this._logger.debug(e.stack);
                }
            }
            
            delete this._parts[partName];
        }
    },
    
    _findDefaultPreset: function BarApp__findDefaultPreset() {
        this._logger.config("Looking for default preset...");
        
        let defaultPresetUrlPrefPath = this.name + this._consts.PREF_DEFAULT_PRESET_URL;
        let internalPresetPath = "$custombar/presets/" + this._consts.DEF_PRESET_FILE_NAME;
        let preferences = this.core.Lib.Preferences;
        let strutils = this.core.Lib.strutils;
        let fileutils = this._barCore.Lib.fileutils;
        let presetFile;
        try {
            let presetUrl = preferences.get(defaultPresetUrlPrefPath, null);
            
            if (!presetUrl)
                throw new Error("Can't get default preset preference value.");
            
            let presetFileName = encodeURIComponent(presetUrl);
            
            presetFile = this.directories.presetsDir;
            presetFile.append(presetFileName);
            
            return this._defaultPreset = new this.BarPlatform.Preset(presetFile, presetUrl);
        }
        catch (e) {
            let presetFilePath = presetFile ? presetFile.path : "no file";
            this._logger.debug( strutils.formatString(
                "Failed parsing normal default preset (%1).\n %2", [presetFilePath, strutils.formatError(e)]) );
            
            if (presetFile)
                fileutils.removeFileSafe(presetFile);
            let extPresetFile = this.directories.presetsDir;
            extPresetFile.append(this._consts.DEF_PRESET_FILE_NAME);
            try {
                let preset = new this.BarPlatform.Preset(extPresetFile);
                try {
                    extPresetFile.moveTo(null, encodeURIComponent(preset.url));
                    preferences.set(defaultPresetUrlPrefPath, preset.url);
                }
                catch (e) {
                    this._logger.error("Could not set external default preset as active. " +
                                       this._barCore.Lib.strutils.formatError(e));
                }
                return this._defaultPreset = preset;
            }
            catch (e) {
                if (extPresetFile.exists()) {
                    this._logger.debug("Failed parsing external default preset.\n" +
                                       this._barCore.Lib.strutils.formatError(e));
                    fileutils.removeFileSafe(extPresetFile);
                }
                try {
                    let presetDoc = fileutils.xmlDocFromStream(this.addonFS.getStreamFromPath(internalPresetPath));
                    let preset = new this.BarPlatform.Preset(presetDoc);
                    
                    try {
                        let destFileName = encodeURIComponent(preset.url);
                        let fMask = parseInt("0755", 8);
                        this.addonFS.copySource(internalPresetPath, this.directories.presetsDir, destFileName, fMask);
                        preferences.set(defaultPresetUrlPrefPath, preset.url);
                    }
                    catch (e) {
                        this._logger.error("Could not extract internal preset.\n" +
                                           this._barCore.Lib.strutils.formatError(e));
                    }
                    
                    this._usingInternalPreset = true;
                    return this._defaultPreset = preset;
                }
                catch (e) {
                    this._logger.fatal("Failed parsing internal default preset.\n" +
                                       this._barCore.Lib.strutils.formatError(e));
                    this._logger.debug(e.stack);
                }
            }
        }
        
        return null;
    },
    _cleanupPreferences: function BarApp__clearPreferences() {
        let currentSetData = this.overlayProvider.currentSetIds;
        if (this._barCore.Lib.sysutils.isEmptyObject(currentSetData))
            return;
        
        let prefService = Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefService);
        let settingKeyPattern = /^(.+#.+)\.(\d+)\..+$/;
        
        function checkBranch(prefBranch) {
            for each (let key in prefBranch.getChildList("", {})) {
                let keyMatch = key.match(settingKeyPattern);
                if (!keyMatch)
                    continue;
                
                let prefProtoID = keyMatch[1];
                let prefInstID = keyMatch[2];
                let instArray = currentSetData[prefProtoID];
                if (!instArray) {
                    prefBranch.deleteBranch(prefProtoID);
                }
                else {
                    if (instArray.indexOf(prefInstID) < 0) {
                        let settingKey = prefProtoID + "." + prefInstID;
                        prefBranch.deleteBranch(settingKey);
                    }
                }
            }
        }
        
        checkBranch( prefService.getBranch(this._barCore.xbWidgetsPrefsPath) );
        checkBranch( prefService.getBranch(this._barCore.nativesPrefsPath) );
    },
    
    
    _openWindow: function BarApp__openWindow(navigatorWindow, path, windowClass, focusIfOpened,
                                             resizeable, modal, windowArgs) {
        let baseNameMatch = path.match(/(\w+)\.x[um]l$/i);
        windowClass = windowClass || (this.name + baseNameMatch? (":" + baseNameMatch[0]): "");
        
        if (focusIfOpened) {
            let chromeWindow = this._barCore.Lib.misc.getTopWindowOfType(windowClass);
            if (chromeWindow) {
                chromeWindow.focus();
                return chromeWindow;
            }
        }
        
        let features = ["chrome", "titlebar", "toolbar", "centerscreen", (modal ? "modal" : "dialog=no")];
        if (resizeable)
            features.push("resizable");
        
        let ownerWindow = navigatorWindow || this._barCore.Lib.misc.getTopBrowserWindow();
        let openParams = [path, windowClass, features.join()].concat(windowArgs);
        
        return ownerWindow.openDialog.apply(ownerWindow, openParams);
    }
};
