"use strict";

const EXPORTED_SYMBOLS = ["overlayProvider"];

const { classes: Cc, interfaces: Ci, utils: Cu } = Components;



const overlayProvider = {
    init: function Overlay_init(barApplication) {
        this._application = barApplication;
        this._barCore = barApplication.core;
        this._logger = barApplication.getLogger("XULOverlay");
        this._commonItemPattern = new RegExp("^" + this._application.name + "\\.cb\\-(\\S+)\\-inst\\-(.+)$");
        this._defaultItemPattern = new RegExp("^" + this._application.name + "\\.cb\\-default\\-(\\d+)$");
        this._defaultSetIDs = {};
        this._unicodeConverter = Cc["@mozilla.org/intl/scriptableunicodeconverter"].createInstance(Ci.nsIScriptableUnicodeConverter);
        this._unicodeConverter.charset = "UTF-8";
        this._DOMSerializer = Cc["@mozilla.org/xmlextras/xmlserializer;1"].getService(Ci.nsIDOMSerializer);
        this._overlayDocString = '\
            <!DOCTYPE overlay [ \
                <!ENTITY app.name "' + this._application.name + '"> \
                <!ENTITY bar.name "' + this._application.branding.expandBrandTemplates("{product1.nom}") + '"> \
            ]> \
            <overlay xmlns="http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul"> \
              <script type="application/x-javascript" src="chrome://&app.name;/content/custombar/overlay/browser.js"/> \
              <toolbarpalette id="BrowserToolbarPalette"/> \
              <toolbox id="navigator-toolbox"> \
                <toolbar id="&app.name;-bar" \
                         cb-app="&app.name;" \
                         toolbarname="&bar.name;" \
                         class="chromeclass-toolbar" \
                         context="toolbar-context-menu" \
                         hidden="false" \
                         mode="full" iconsize="small" defaulticonsize="small" lockiconsize="true" \
                         customizable="true" \
                         crop="end" \
                         defaultset=""/> \
              </toolbox> \
            </overlay>';
        try {
            let swMapStream = this._application.addonFS.getStreamFromPath(this._consts.STR_SWMAP_PATH);
            this._staticWidgets = JSON.parse( this._application.core.Lib.fileutils.readStringFromStream(swMapStream) );
        }
        catch (e) {
            this._logger.error("Could not load static widgets info. " + this._barCore.Lib.strutils.formatError(e));
        }
        
        Cc["@mozilla.org/network/protocol;1?name=" + this._application.name]
            .getService(Ci.nsIProtocolHandler)
            .wrappedJSObject
            .addDataProvider(this);
    },
    
    finalize: function Overlay_finalize() {
        this._overlayDoc = null;
    },
    newURI: function Overlay_newURI(aSpec, aOriginalCharset, aBaseURI) {
        if (!aSpec)
            return null;

        let content = "";
        try {
            switch (aSpec.toLowerCase()) {
                case "browser-overlay":
                    content = btoa(this._getBrowserOverlay());
                    break;

                case "browser-yafeatures-overlay":
                    content = btoa(this._getBrowserStaticOverlay());
                    break;

                default:
                    return null;
            }
        }
        catch (e) {
            this._logger.error(this._application.core.Lib.strutils.formatError(e));
            this._logger.debug(e.stack);
        }

        let uri =  Cc["@mozilla.org/network/simple-uri;1"].createInstance(Ci.nsIURI);
        uri.spec = "data:application/vnd.mozilla.xul+xml;base64," + content;
        return uri;
    },
    
    get currentSetIds() {
        if (!this._currentSetIds)
            this._currentSetIds = this._readCurrentSetIds();
        return this._currentSetIds;
    },
    
    get staticWidgets() {
        return this._barCore.Lib.sysutils.copyObj(this._staticWidgets);
    },
    
    parseWidgetItemId: function Overlay_parseWidgetItemId(itemID, fullMode) {
        let match;
        
        if (match = itemID.match(this._commonItemPattern))
            return {
                prototypeID: match[1],
                instanceID: match[2],
                isFromDefaultSet: false
            };
        
        if (match = itemID.match(this._defaultItemPattern)) {
            let presetItemIndex = parseInt(match[1], 10);
            let widgetEntry = this._application.defaultPreset.visibleWidgetEntries[presetItemIndex];
            if (widgetEntry) {
                let itemInfo = {
                    prototypeID: widgetEntry.componentID,
                    instanceID: 0,
                    isFromDefaultSet: true
                };
                
                if (fullMode) {
                    let instID;
                    if (presetItemIndex in this._defaultSetIDs)
                        itemInfo.instanceID = this._defaultSetIDs[presetItemIndex];
                    else
                        itemInfo.instanceID = this._defaultSetIDs[presetItemIndex] = this._application.getNewWidgetInstanceID();
                    itemInfo.settings = this._application.core.Lib.sysutils.copyObj(widgetEntry.settings);
                }
                return itemInfo;
            }
        }
        
        return null;
    },
    
    compileWidgetItemId: function Overlay_compileWidgetItemId(protoID, instanceID) {
        return this._application.name + ".cb-" + protoID + "-inst-" + instanceID;
    },
    
    widgetItemRemoved: function Overlay_widgetItemRemoved(instID) {
        for (let key in this._defaultSetIDs) {
            if (this._defaultSetIDs[key] == instID) {
                delete this._defaultSetIDs[key];
            }
        }
    },
    makePaletteItem: function Overlay_makePaletteItem(doc, widgetInfo, instanceID) {
        let toolbarItem = doc.createElementNS(this._consts.STR_XUL_NS, "toolbaritem");
        
        toolbarItem.setAttribute("id", this.compileWidgetItemId(widgetInfo.id, instanceID));
        toolbarItem.setAttribute("cb-proto-id", widgetInfo.id);
        toolbarItem.setAttribute("cb-app", this._application.name);
        toolbarItem.setAttribute("title", widgetInfo.name);
        toolbarItem.setAttribute("image", widgetInfo.iconPath ? widgetInfo.package_.resolvePath(widgetInfo.iconPath) : "");
        
        return toolbarItem;
    },
    
    makeDefaultSetItems: function Overlay_makeDefaultSetItems(document) {
        return this._makeDefaultSetItems(document);
    },
    reloadStaticWidgets: function Overlay_reloadStaticWidgets(aDocument) {
        let palette = aDocument.getElementById("navigator-toolbox").palette;
        
        if (!palette)
            return;
        let widgetsIds = [
            "yasearch-urlbar-icon",
            "yasearchQSendToMail", "yasearchQSendToYaru",
            "yaBookmarksPopup",
            "yasearch-cy", "yasearch-spam", "yasearch-bloggers", "yasearch-spellchecker",
            "yasearch-translate", "yasearch-login", "yasearch-mail", "yasearch-fotki",
            "yasearch-yaru", "yasearch-moikrug", "yasearch-lenta", "yasearch-money",
            "yasearch-bookmarks", "yasearch-preferences"
        ];
        
        let doc = this._getBrowserStaticOverlayDocument();
        
        for (let i = 0, len = widgetsIds.length; i < len; i++) {
            let id = widgetsIds[i];
            let element = doc.getElementById(id);
            if (!element || aDocument.getElementById(id))
                continue;
            
            let parent = element.parentNode;
            
            if (parent.id == "BrowserToolbarPalette") {
                if (!palette.getElementsByAttribute("id", id).length)
                    palette.appendChild(element.cloneNode(true));
            } else {
                parent = aDocument.getElementById(parent.id);
                if (parent)
                    parent.appendChild(element.cloneNode(true));
            }
        }
    },
    _staticWidgets: {},
    _commonItemPattern: null,
    _defaultItemPattern: null,
    _defaultSetIDs: null,
    _unicodeConverter: null,
    _DOMSerializer: null,
    _overlayDocString: undefined,
    _overlayDoc: null,
    _logger: null,
    
    _consts: {
        STR_SWMAP_PATH: "$custombar/native_widgets.json",
        STR_XUL_NS: "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul",
        ERR_CREATE_ITEM: "Could not add widget palette item"
    },
    
    _getBrowserOverlay: function Overlay__getBrowserOverlay() {
            this._overlayDoc = this._createBrowserOverlay();
        return this._unicodeConverter.ConvertFromUnicode(this._DOMSerializer.serializeToString(this._overlayDoc));
    },
    
    _createBrowserOverlay: function Overlay__createBrowserOverlay() {
        let start = Date.now();
        
        let DOMParser = this._application.core.Lib.xmlutils.getDOMParser(null, null, true);
        
        let overlayDoc = DOMParser.parseFromString(this._overlayDocString, "text/xml"),
            toolbarPalette = overlayDoc.getElementById("BrowserToolbarPalette"),
            toolbar = overlayDoc.getElementById(this._application.name + "-bar");
        
        let [defaultSetItems, defaultSetIDs] = this._makeDefaultSetItems(overlayDoc);
        for (let i = 0, len = defaultSetItems.length; i < len; i++)
            toolbarPalette.appendChild(defaultSetItems[i]);
        
        toolbar.setAttribute("defaultset", defaultSetIDs.join(","));
        
        let paletteItems = this._makePaletteItems(overlayDoc);
        for (let i = 0, len = paletteItems.length; i < len; i++)
            toolbarPalette.appendChild(paletteItems[i]);
        
        this._logger.debug("Overlay created in " + (Date.now() - start) + "ms");
        
        return overlayDoc;
    },
    
    _makePaletteItems: function Overlay__makePaletteItems(overlayDoc) {
        let currentSetIDsData = this._readCurrentSetIds(),
            widgetLibrary = this._application.widgetLibrary,
            paletteItems = [];
        let widgetsInfo = widgetLibrary.getWidgetsInfo();
        for (let i = 0, len = widgetsInfo.length; i < len; i++) {
            let widgetInfo = widgetsInfo[i];
            let isUsed = false;
            
            let protoInstHash = currentSetIDsData[widgetInfo.id] || null;
            if (protoInstHash) {
                isUsed = true;
                for each (let instID in protoInstHash) {
                    paletteItems.push(this.makePaletteItem(overlayDoc, widgetInfo, instID));
                    this._application.getNewWidgetInstanceID();
                    
                    if (widgetInfo.isUnique)
                        break;
                }
            }
            
            if (!widgetInfo.isUnique || !isUsed) {
                paletteItems.push(this.makePaletteItem(overlayDoc, widgetInfo, 0));
            }
        }
        
        return paletteItems;
    },
    
    _makeDefaultSetItems: function Overlay__makeDefaultSetItems(overlayDoc) {
        let widgetLibrary = this._application.widgetLibrary,
            paletteItems = [],
            defaultSetIDs = [];
        
        let widgetEntries = this._application.defaultPreset.visibleWidgetEntries,
            usedIDs = { __proto__: null };
        
        for (let i = 0, length = widgetEntries.length; i < length; i++) {
            let widgetEntry = widgetEntries[i],
                widgetInfo = null,
                protoID = widgetEntry.componentID,
                itemID;
            
            if (protoID in this._staticWidgets) {
                itemID = this._staticWidgets[protoID];
            } else {
                try {
                    widgetInfo = widgetLibrary.getWidgetInfo(protoID);
                    if (widgetInfo.isUnique && (protoID in usedIDs))
                        continue;
                }
                catch (e) {
                    this._logger.error(this._consts.ERR_CREATE_ITEM + ". " +
                                       this._application.core.Lib.strutils.formatError(e));
                    continue;
                }
                
                let paletteItem = this.makePaletteItem(overlayDoc, widgetInfo, 0);
                itemID = this._application.name + ".cb-default-" + i;
                paletteItem.setAttribute("id", itemID);
                paletteItem.setAttribute("cb-defaultset", "");
                paletteItems.push(paletteItem);
                
                usedIDs[protoID] = true;
            }
            
            defaultSetIDs.push(itemID);
        }
        
        return [paletteItems, defaultSetIDs];
    },
    _readCurrentSetIds: function Overlay__readCurrentSetIds() {
        let result = {};
        
        let commonItemPattern = this._commonItemPattern;
        
        function _getIdsFromCurrentset(aCurrenSetString) {
            if (!aCurrenSetString)
                return;
            
            let currentSetIds = aCurrenSetString.split(",");
            for (let i = 0, len = currentSetIds.length; i < len ; i++) {
                let barWidgetIDMatch = currentSetIds[i].match(commonItemPattern);
                if (barWidgetIDMatch) {
                    let widgetProtoID = barWidgetIDMatch[1];
                    let widgetInstance = barWidgetIDMatch[2];
                    let instArray = result[widgetProtoID];
                    if (!instArray) {
                        instArray = [];
                        result[widgetProtoID] = instArray;
                    }
                    instArray.push(widgetInstance);
                }
            }
        }
        
        let rdfService = Cc["@mozilla.org/rdf/rdf-service;1"].getService(Ci.nsIRDFService);
        let localStoreDataSource = rdfService.GetDataSource("rdf:local-store");
        
        let allResources = localStoreDataSource.GetAllResources();
        let currentSetResource = rdfService.GetResource("currentset");
        
        while (allResources.hasMoreElements()) {
            let res = allResources.getNext().QueryInterface(Ci.nsIRDFResource);
            let tool = res.Value;
            
            if (tool) {
                if (tool == "chrome://browser/content/browser.xul#customToolbars") {
                    let customToolbarsResource = rdfService.GetResource(tool);
                    let index = 0;
                    let currentSetTarget;
                    do {
                        let toolbarResource = rdfService.GetResource("toolbar" + (++index));
                        currentSetTarget = localStoreDataSource.GetTarget(customToolbarsResource, toolbarResource, true);
                        if (currentSetTarget instanceof Ci.nsIRDFLiteral) {
                            let ids = currentSetTarget.Value.split(":");
                            ids.shift();
                            _getIdsFromCurrentset(ids.join(":"));
                        }
                    } while (currentSetTarget);
                } else {
                    let toolbarResource = rdfService.GetResource(tool);
                    let currentSetTarget = localStoreDataSource.GetTarget(toolbarResource, currentSetResource, true);
                    if (currentSetTarget instanceof Ci.nsIRDFLiteral)
                        _getIdsFromCurrentset(currentSetTarget.Value);
                }
            }
        }
        
        return result;
    },
    
    _getBrowserStaticOverlayDocument: function Overlay__getBrowserStaticOverlayDocument() {
        return this._application.core.Lib.xmlutils.getDOMParser(null, null, true)
                   .parseFromString(this._getBrowserStaticOverlay(), "text/xml");
    },
    _getBrowserStaticOverlay: function Overlay__getBrowserStaticOverlay() {
        let branding = this._application.branding;
        
        let defaultWidgetsIDs = { __proto__: null };
        this._application.defaultPreset.widgetEntries
            .forEach(function(widgetEntry) {
                defaultWidgetsIDs[widgetEntry.componentID] = true;
            });
        
        let scope = {
            content: "",
            _staticWidgets: this.staticWidgets,
            _defaultWidgetsIDs: defaultWidgetsIDs,
            hasWidget: function StaticOverlayScope_hasWidget(aWidgetId) {
                let id = "http://bar.yandex.ru/packages/yandexbar#" + aWidgetId;
                return (id in this._staticWidgets) && (id in this._defaultWidgetsIDs);
            },
            hasFeature: function StaticOverlayScope_hasFeature(aFeatureName) {
                return branding.getYandexFeatureState(aFeatureName);
            }
        };
        
        Cc["@mozilla.org/moz/jssubscript-loader;1"]
            .getService(Ci.mozIJSSubScriptLoader)
            .loadSubScript("chrome://" + this._application.name + "/content/custombar/overlay/" +
                           "browser-static-widgets.xul.js", scope);
        return scope.content.replace(/>\s+</g, "><");
    }
};