"use strict";

const EXPORTED_SYMBOLS = ["BarPlatform"];
const PLATFORM_MODULES = [
    "preset.js", "manifest.js", "preset-with-manifest.js", "package.js", "unit.js", "permissions.js", "blacklist.js"
];

const { classes: Cc, interfaces: Ci, utils: Cu } = Components;

Cu.import("resource://gre/modules/XPCOMUtils.jsm");

var sysutils,
    fileutils,
    strutils,
    xmlutils,
    netutils,
    misc,
    CustomErrors,
    Preferences,
    barApp,
    Base;

const BarPlatform = {
    init: function BarPlatform_init(application) {
        barApp = this._application = application;
        
        let (appCoreLib = application.core.Lib) {
            sysutils = appCoreLib.sysutils;
            fileutils = appCoreLib.fileutils;
            strutils = appCoreLib.strutils;
            xmlutils = appCoreLib.xmlutils;
            netutils = appCoreLib.netutils;
            misc = appCoreLib.misc;
            CustomErrors = appCoreLib.CustomErrors;
            Preferences = appCoreLib.Preferences;
            Base = appCoreLib.Base;
        };
        
        this._loggersRoot = application.name + ".CP";
        this._logger = application.core.Lib.Log4Moz.repository.getLogger(this._loggersRoot);
        this._loadModules();
    },
    
    finalize: function BarPlatform_finalize() {
        this._parsers = null;
        this._logger = null;
    },
    
    registerUnitParser: function BarPlatform_registerUnitParser(platformType, componentType, parser) {
        let parserKey = platformType + "#" + componentType;
        if (!parser)
            delete this._parsers[parserKey];
        else
            this._parsers[parserKey] = parser;
    },
    
    
    findPlatformMatch: function BarPlatform_findPlatformMatch(items) {
        if (!sysutils.isArray(items))
            throw new CustomErrors.EArgRange("items", "Array", items);
        
        let platformInfo = sysutils.platformInfo;
        for (let i = 0, len = items.length; i < len; i++) {
            let item = items[i];
            if (
                (("os" in item) && (item.os != platformInfo.os.name)) ||
                (("browser" in item) && (item.browser != platformInfo.browser.name)) ||
                (("architecture" in item) && (item.architecture != platformInfo.browser.simpleArchitecture))
            )
                continue;
            return item;
        }
        return null;
    },
    
    
    findBestLocalizedElement: function BarPlatform_findBestLocalizedElement(elementsArray) {
        let localizedElements = {};
        for each (let element in elementsArray) {
            let elemLocaleStr = element.getAttribute("locale") || element.getAttribute("lang") || "";
            localizedElements[elemLocaleStr] = element;
        }
        
        return misc.findBestLocalizedValue(localizedElements, this._application.locale);
    },

    parseComponentID: function BarPlatform_parseComponentID(componentID) {
        if (typeof componentID != "string")
            throw new CustomErrors.EArgType("componentID", "string", componentID);
        
        let separatorPos = componentID.indexOf("#");
        let packageID = componentID.substring(0, separatorPos);
        if (!packageID)
            throw new Error(this._consts.ERR_INVALID_COMP_ID + componentID);
        let compName = componentID.substring(separatorPos + 1);
        if (!compName)
            throw new Error(this._consts.ERR_INVALID_COMP_ID + componentID);
        return [packageID, compName];
    },
    
    parsePrefPath: function BarPlatform_parsePrefPath(prefPath, compID) {
        let appCore = this._application.core;
        let [packageID, compName] = this.parseComponentID(compID);
        let partsRE = new RegExp( strutils.escapeRE(appCore.nativesPrefsPath + packageID) + "(\#(.+)\\.(\\w+))?\\.settings\\.(.+)");
        let partsMatch = prefPath.match(partsRE);
        if (!partsMatch)
            throw new Error("Invalid prefPath");
        let isPackagePref = !partsMatch[1],
            scope = partsMatch[3],
            isInstancePref = (!isPackagePref && (scope != "all")),
            settingName = partsMatch[4];
        let prefProps = {
            settingName: settingName,
            isPackagePref: isPackagePref,
            isComponentPref: !(isPackagePref || isInstancePref),
            isInstancePref: isInstancePref,
            instanceID: isInstancePref ? scope : undefined
        };
        return prefProps;
    },
    
    eraseSettings: function BarPlatform_eraseSettings() {
        let barCore = this._application.core;
        this._obsService.notifyObservers(null, barCore.eventTopics.EVT_BEFORE_GLOBAL_RESET, null);
        try {
            Preferences.resetBranch(barCore.xbWidgetsPrefsPath);
            Preferences.resetBranch(barCore.nativesPrefsPath);
            
            fileutils.removeFileSafe(this._application.directories.nativeStorageDir);
            let dataStream = this._application.addonFS.getStreamFromPath(this._consts.STR_STATIC_PREFS_PATH);
            let staticPrefsInfo = JSON.parse(fileutils.readStringFromStream(dataStream));
            const APP_NAME = this._application.name;
            for each (let [, prefPath] in Iterator(staticPrefsInfo.preferences))
                Preferences.reset(APP_NAME + prefPath);
            for each (let [, branchPath] in Iterator(staticPrefsInfo.branches))
                Preferences.resetBranch(APP_NAME + branchPath);
        }
        finally {
            this._obsService.notifyObservers(null, barCore.eventTopics.EVT_AFTER_GLOBAL_RESET, null);
        }
    },
    
    _parsers: {},
    _application: null,
    _loggersRoot: undefined,
    _logger: null,
    get _obsService() {
        let obs = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
        delete this._obsService;
        return this._obsService = obs;
    },
    _consts: {
        STR_STATIC_PREFS_PATH: "$custombar/preferences/static-widgets/resetprefs.json",
        ERR_INVALID_COMP_ID: "Invalid component ID "
    },
    
    _loadModules: function BarPlatform__loadModules() {
        const mozSSLoader = Cc["@mozilla.org/moz/jssubscript-loader;1"].getService(Ci.mozIJSSubScriptLoader);
        const xbDirPath = this._application.partsURL + "platform/";
        PLATFORM_MODULES.forEach(function BarPlatform_loadModule(moduleFileName) {
            this._logger.debug("  Loading module " + moduleFileName);
            mozSSLoader.loadSubScript(xbDirPath + moduleFileName);
        }, this);
    },
    
    _getLogger: function BarPlatform__getLogger(name) {
        return this._application.core.Lib.Log4Moz.repository.getLogger(this._loggersRoot + "." + name);
    },
    
    _getParser: function BarPlatform__getParser(platformType, componentType) {
        let parser = this._parsers[platformType + "#" + componentType];
        if (!parser)
            throw new Error(strutils.formatString("No \"%2\" parser registered for API \"%1\".",
                                                  [platformType, componentType]));
        return parser;
    }
};


