"use strict";



BarPlatform.FilePackage = function FilePackage(rootDir , id , permissions , domain ) {
    if (!(rootDir instanceof Ci.nsIFile))
        throw new CustomErrors.EArgType("rootDir", "nsIFile", rootDir);
    if (!rootDir.isDirectory())
        throw new CustomErrors.EArgRange("rootDir", "nsIFile(Directory)", rootDir);
    if (!sysutils.isString(id))
        throw new CustomErrors.EArgType("id", "String", id);
    
    this._rootDir = rootDir.clone();
    this._rootDir.normalize();
    
    this._id = id;
    if (domain) {
        this._domain = domain;
    }
    else {
        let uuid = Cc["@mozilla.org/uuid-generator;1"].getService(Ci.nsIUUIDGenerator).generateUUID().toString();
        this._domain = uuid.replace(/[^\w\-]/g, "") + "." + barApp.name;
    }
    
    this._files = {__proto__: null};
    this._logger = BarPlatform._getLogger("Package_" + this._domain);
    
    let protocolHandler = barApp.core.xbProtocol;
    protocolHandler.setDataProvider(this._domain, this);
    this._uri = protocolHandler.newURI(protocolHandler.scheme + "://" + this._domain + "/", null, null);
    
    this._persistPath = barApp.core.xbWidgetsPrefsPath + this._id;
    this._settings = {};
    this._units = {};
    this._permissions = permissions;
};

BarPlatform.FilePackage.prototype = {
    constructor: BarPlatform.FilePackage,
    get id() {
        return this._id;
    },
    get rootDirectory() {
        return this._rootDir.clone();
    },
    
    get permissions() {
        return this._permissions;
    },
    getUnit: function FilePkg_getUnit(unitName) {
        let unit = this._units[unitName];
        if (!unit) {
            let unitFile = this.getFile(unitName + ".xml");
            
            unit = new BarPlatform.Unit(unitFile.leafName, this, unitName);
            this._units[unitName] = unit;
        }
        return unit;
    },
    
    newChannelFromPath: function FilePkg_newChannelFromPath(path) {
        return this.newChannel(this._ioService.newURI(this.resolvePath(path), null, null));
    },
    
    
    getXMLDocument: function FilePkg_getXMLDocument(path, usePrivilegedParser) {
        let channel = this.newChannelFromPath(path);
        return fileutils.xmlDocFromStream(channel.open(), channel.originalURI, channel.originalURI, usePrivilegedParser);
    },
    resolvePath: function FilePkg_resolvePath(path , base ) {
        if (!sysutils.isString(path))
            throw new CustomErrors.EArgType("path", "String", path);
        
        if (base) path = path.replace(/^(?!\/|\w+:)/, base.replace(/[^\/]+$/, ''));
        
        return this._uri.resolve(path);
    },
    
    finalize: function FilePkg_finalize() {
        for (let unitName in this._units)
            try {
                this._units[unitName].finalize();
            }
            catch (e) {
                this._logger.error("Couldn't clear loaded unit " + unitName + ". " + strutils.formatError(e));
            }
        this._units = null;
        this._settings = null;
        
        barApp.core.xbProtocol.setDataProvider(this._domain, null);
    },
    addSettingUser: function FilePkg_addSettingUser(settingName, componentID) {
        let users = this._settings[settingName] || (this._settings[settingName] = []);
        if (users.indexOf(componentID) > -1) return;
        users.push(componentID);
    },
    
    removeSettingUser: function FilePkg_removeSettingUser(settingName, componentID) {
        if (!settingName in this._settings) return;
        let users = this._settings[settingName];
        let userIndex = users.indexOf(componentID);
        if (userIndex < 0) return;
        users.splice(userIndex, 1);
    },
    
    getSettingUsers: function FilePkg_getSettingUsers(settingName) {
        let users = this._settings[settingName];
        return users ? users.slice(0) : [];
    },
    QueryInterface: XPCOMUtils.generateQI([Ci.nsISupports, Ci.nsIXBDataProvider, Ci.nsIXBPackage]),
    
    get wrappedJSObject() {
        return this;
    },
    
    get UUID() {
        return this._domain;
    },
    
    get _chromeChannelPrincipal() {
        let chromeChannel = this._ioService.newChannel("chrome://" + barApp.name + "/content/", null, null);
        let systemPrincipal = chromeChannel.owner;
        chromeChannel.cancel(Components.results.NS_BINDING_ABORTED);
        
        this.__defineGetter__("_chromeChannelPrincipal", function() systemPrincipal);
        return this._chromeChannelPrincipal;
    },
    
    newChannel: function FilePkg_newChannel(aURI) {
        let file = this.getFile(aURI.QueryInterface(Ci.nsIURL).filePath);
        
        let filesStream = Cc["@mozilla.org/network/file-input-stream;1"]
            .createInstance(Ci.nsIFileInputStream);
        filesStream.init(file, 1 , 0, filesStream.CLOSE_ON_EOF);
        
        let channel = Cc["@mozilla.org/network/input-stream-channel;1"]
                          .createInstance(Ci.nsIInputStreamChannel)
                          .QueryInterface(Ci.nsIChannel);
        channel.setURI(aURI);
        channel.originalURI = aURI;
        channel.contentStream = filesStream;
        channel.owner = this._chromeChannelPrincipal;
        
        return channel;
    },
    
    getFile: function FilePkg_getFile(path) {
        let file = this.findFile(path);
        if (!file)
            throw new Error(this._consts.ERR_FILE_NOT_FOUND + " \"" + path + "\"");
        return file;
    },
    findFile: function FilePkg_findFile(path) {
        if (!sysutils.isString(path))
            throw new CustomErrors.EArgType("path", "String", path);
        
        path = this._suppressRelativePathReference(path);
        
        if (path in this._files)
            return this._files[path];
        
        let components = path.split("/");
        if (components[components.length - 1][0] == ".")
            return this._files[path] = null;
        
        let root = this._rootDir.clone();
        let locales = this._locales();
        let file = null;
        
        for (let i = locales.length; i--;) {
            let localeName = locales[i].name;
            let candidate = root.clone();
            if (localeName != "") {
                candidate.append("locale");
                candidate.append(localeName);
            }
            
            for (let j = 0, len = components.length; j < len; j++)
                candidate.append(components[j]);
            
            if (!candidate.exists())
                continue;
            
            try {
                candidate.normalize();
                if (candidate.isReadable() && this._rootDir.contains(candidate, true)) {
                    file = candidate;
                    break;
                }
            } catch(e) {
                this._logger.error("Error while searching file. " + strutils.formatError(e));
                continue;
            }
        }
        
        return this._files[path] = file;
    },
    
    _consts: {
        ERR_NO_UNIT: "No such unit",
        ERR_NO_WIDGET: "Unit does not describe a widget",
        ERR_FILE_NOT_FOUND: "File not found",
        ERR_ACCESS_DENIED: "Attempt to access a file outside the package directory"
    },
    _id: undefined,
    _domain: undefined,
    _rootDir: null,
    _units: null,
    _logRoot: undefined,
    _name: undefined,
    _bestLocaleName: undefined,
    _logger: null,
    _persistPath: undefined,
    
    _ioService: Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService),
    
    _locales: function FilePkg__locales() {
        if (this._localesCache)
            return this._localesCache;
        
        const weights = {
            language: 32,
            root: 16,
            ru: 8,
            en: 4,
            country: 2,
            region: 1
        };
        
        let locales = [];
        locales.push({
            name: "",
            weight: weights.root,
            components: null
        });
        
        let localeDir = this._rootDir.clone();
        localeDir.append("locale");
        if (!localeDir.exists())
            return this._localesCache = locales;
        
        let appLocale = misc.parseLocale(barApp.localeString);
        
        let entries = localeDir.directoryEntries;
        while (entries.hasMoreElements()) {
            let file = entries.getNext().QueryInterface(Ci.nsIFile);
            if (file.isDirectory()) {
                let name = file.leafName;
                let components = misc.parseLocale(name);
                
                if (!components)
                    continue;
                
                let weight = 0;
                for (let space in weights) {
                    let component = components[space];
                    if (component === undefined)
                        continue;
                    
                    if (space == "language")
                        if (component in weights)
                            weight += weights[component];
                    
                    if (component === appLocale[space])
                        weight += weights[space];
                }
                
                locales.push({
                    name: name,
                    weight: weight,
                    components: components
                });
                
            }
        }
        
        locales.sort(function FilePkg__locales_sort(a, b) a.weight - b.weight);
        
        return this._localesCache = locales;
    },
    
    _suppressRelativePathReference: function FilePkg__suppressRelativePathReference(path) {
        let re = [
            /\/\//g,
            /\/\.\//g,
            /\/[^\/]+\/\.\.\//g,
            /\/\.\.\//g
        ];
        
        for (let i = 0, len = re.length; i < len; i++)
            while (re[i].test(path))
                path = path.replace(re[i], "/");
        
        return path;
    }
};
