"use strict";

const EXPORTED_SYMBOLS = ["updater"];

const { classes: Cc, interfaces: Ci, utils: Cu, results: Cr } = Components;
var sysutils,
    fileutils,
    strutils,
    netutils,
    misc,
    CustomErrors,
    Preferences,
    barApplication;


const defaultController = {
    start: function DefaultController_start(updateProcess) {
        
        if (updater._updatesThrottled) {
            this._logger.debug("Throttling user components update check");
            this._silentOnly = true;
            
            updater._updatesThrottlingTick();
        }
        
        this._updateProcess = updateProcess;
        this._updateProcess.checkPresets(this._silentOnly);
    },
    
    onPresetsDownloaded: function DefaultController_onPresetsDownloaded(newDefPresetPair, newPresetPairs) {
        this._newPresetPairs = newPresetPairs;
        if (newDefPresetPair)
            this._updateProcess.startSilentDefaultPresetUpdate();
        else if (!this._silentOnly)
            this._startExplicit();
    },
    
    onDefaultPresetUpdateFinish: function DefaultController_onDefaultPresetUpdateFinish(checkedPkgIDsSet) {
        if (!this._silentOnly)
            this._startExplicit();
    },
    
    onExplicitUpdateDataReady: function DefaultController_onExplicitUpdateDataReady(newPresets, newManifests) {
        if (!newPresets.length && !newManifests.length) {
            this._logger.debug("No new components found. Update check finished.");
            return;
        }
        this._logger.debug(strutils.formatString("Found %1 new presets and %2 new manifests.",
                                                 [newPresets.length, newManifests.length]));
        
        this._newUserPresets = newPresets;
        this._newUserManifests = newManifests;
        
        this._askUserForUpdateComponents();
    },
    
    onUpdateFinished: function DefaultController_onUpdateFinished(somethingUpdated) {
        if (somethingUpdated)
            updater.resetUpdatesThrottling();
    },
    get _logger() {
        delete this._logger;
        return this._logger = barApplication.getLogger("DefUpdCtrl");
    },
    _newPresetPairs: [],
    _silentOnly: false,
    
    _startExplicit: function DefaultController__startExplicit() {
        this._updateProcess.startUserComponentsUpdate(this._newPresetPairs);
    },
    
    _askUserForUpdateComponents: function DefaultController__askUserForUpdateComponents() {
        updater._updatesThrottlingTick();
        
        misc.openWindow({
            url: "chrome://" + barApplication.name + "/content/custombar/dialogs/package-management/notification-update/notification-update.xul",
            features: "__popup__",
            name: "notification-update",
            presets: this._newUserPresets,
            manifests: this._newUserManifests,
            application: barApplication,
            then: sysutils.bindFunction(this._onUserAnswer, this)
        });
    },
    
    _onUserAnswer: function DefaultController__onUserAnswer(install, fastInstall) {
        if (!install) return;
        
        misc.openWindow({
            url: "chrome://" + barApplication.name + "/content/custombar/dialogs/package-management/update/update.xul",
            features: "__popup__",
            name: "package-management-update",
            mode: "update",
            presets: this._newUserPresets,
            manifests: this._newUserManifests,
            application: barApplication,
            simple: fastInstall,
            then: sysutils.bindFunction(this.onUpdateFinished, this)
        });
    }
};

const updater = {
    init: function Updater_init(application) {
        barApplication = this._application = application;
        
        let (appCoreLib = application.core.Lib) {
            sysutils = appCoreLib.sysutils;
            fileutils = appCoreLib.fileutils;
            strutils = appCoreLib.strutils;
            netutils = appCoreLib.netutils;
            misc = appCoreLib.misc;
            CustomErrors = appCoreLib.CustomErrors;
            Preferences = appCoreLib.Preferences;
        };
        
        this._logger = application.getLogger("CompUpdater");
        this._timer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
        
        this._timer.initWithCallback(this, this._nextCheckInterval, this._timer.TYPE_ONE_SHOT);
        
        let skippedUpdatesCount = Math.floor((Date.now() - this._lastUpdateTimestamp) / this._updateInterval) - 1;
        if (skippedUpdatesCount < 0) skippedUpdatesCount = 0;
        this.riseSkippedUpdatesThrottling(skippedUpdatesCount);
    },
    
    finalize: function Updater_finalize() {
        this._timer.cancel();
    },
    
    
    checkUpdates: function Updater_checkUpdates(updateController) {
        this._startUpdates(updateController);
    },
    notify: function Updater_notify() {
        try {
            this._startUpdates(defaultController);
        }
        catch (e) {
            this._logger.error("Could not check updates. " + strutils.formatError(e));
            this._logger.debug(e.stack);
        }
    },
    _consts: {
        ERR_PRESET_DL: "Preset %1 was not downloaded. Status code: %2. Error: %3",
        ERR_MFS_DL: "Manifest %1 was not downloaded. Status code: %2. Error: %3",
        ERR_PKG_DL: "Package %1 was not downloaded. Status code: %2. Error: %3",
        ERR_WUNIT_FAIL: "Could not get unit for widget '%1'. %2",
        ERR_PUNIT_FAIL: "Could not get unit for plugin '%1'. %2",
        ERR_W_MISSING: "Widget '%1' is missing",
        ERR_P_MISSING: "Plugin '%1' is missing"
    },
    _comparator: Cc["@mozilla.org/xpcom/version-comparator;1"].getService(Ci.nsIVersionComparator),
    _sysTmpDir: Cc["@mozilla.org/file/directory_service;1"].getService(Ci.nsIProperties).get("TmpD", Ci.nsIFile),
    get _obsService() {
        let obs = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
        delete this._obsService;
        return this._obsService = obs;
    },
    
    get _lastUpdatePrefName() {
        return this._application.name + ".updates.widgets.lastUpdateTime";
    },
    
    get _lastUpdateTimestamp() {
        return parseInt(Preferences.get(this._lastUpdatePrefName, 0), 10) * 1000;
    },
    
    set _lastUpdateTimestamp(val) {
        Preferences.set(this._lastUpdatePrefName, Math.round(val / 1000));
    },
    
    get _updateInterval() {
        return 24 * 60 * 60 * 1000;
    },
    
    get _nextCheckInterval() {
        let nextCheckInterval = this._updateInterval - Math.abs(this._lastUpdateTimestamp - Date.now());
        nextCheckInterval = Math.max(nextCheckInterval, this._application.generateDelay() * 1000);
        return nextCheckInterval;
    },
    
    get _blackListETag() {
        return this._application.preferences.get("blacklist.etag", "");
    },
    
    set _blackListETag(eTag) {
        this._application.preferences.set("blacklist.etag", eTag);
    },
    
    _makeTempFile: function Updater__makeTempFile(URLStr) {
        let fileName;
        try {
            let stdURL = Cc["@mozilla.org/network/standard-url;1"].createInstance(Ci.nsIStandardURL);
            stdURL.init(Ci.nsIStandardURL.URLTYPE_STANDARD, -1, URLStr, null, null);
            let url = stdURL.QueryInterface(Components.interfaces.nsIURL);
            fileName = url.fileName;
        }
        catch (e) {
            this._logger.warn("Could not extract URL fileName. " + e);
        }
        
        let tempFile = this._sysTmpDir.clone();
        tempFile.append(fileName || misc.CryptoHash.getFromString(URLStr, "MD5"));
        tempFile.createUnique(Ci.nsIFile.NORMAL_FILE_TYPE, parseInt("0666", 8));
        return tempFile;
    },
    
    
    _getInstalledPresets: function Updater__getInstalledPresets(defaultOnly) {
        let presets = {};
        
        if (defaultOnly) {
            presets[this._application.defaultPresetURL] = this._application.defaultPreset;
            return presets;
        }
        
        let presetDirEntries = this._application.directories.presetsDir.directoryEntries;
        
        while (presetDirEntries.hasMoreElements()) {
            let file = presetDirEntries.getNext().QueryInterface(Ci.nsIFile);
            try {
                let presetAddress = decodeURIComponent(file.leafName);
                let preset = new this._application.BarPlatform.Preset(file, presetAddress);
                
                let widgetLibrary = this._application.widgetLibrary;
                let used = false;
                for (let componentID in preset.componentIDs) {
                    if (widgetLibrary.isKnownComponent(componentID)) {
                        used = true;
                        break;
                    }
                }
                if (used)
                    presets[preset.address] = preset;
                else
                    file.remove(false);
            }
            catch (e) {
                this._logger.warn("Invalid preset file. " + e);
                this._logger.debug(e.stack);
                file.remove(true);
            }
        }
        
        return presets;
    },
    
    
    _gatherPresetInfo: function Updater__gatherPresetInfo(preset, usedPackageIDs, unknownWIDs, unknownPIDs) {
        let staticWidgets = this._application.overlayProvider.staticWidgets;
        let widgetLibrary = this._application.widgetLibrary;
        
        for each (let presetEntry in preset.allEntries) {
            if (usedPackageIDs)
                usedPackageIDs[presetEntry.packageID] = 1;
            
            if (presetEntry.componentType == presetEntry.TYPE_WIDGET) {
                if (unknownWIDs
                        && !widgetLibrary.isKnownWidget(presetEntry.componentID)
                        && !(presetEntry.componentID in staticWidgets))
                    unknownWIDs[presetEntry.componentID] = 1;
            }
            else {
                if (unknownPIDs && !this._application.widgetLibrary.isKnownPlugin(presetEntry.componentID))
                    unknownPIDs[presetEntry.componentID] = 1;
            }
        }
    },
    
    _startAsyncDownloads: function Updater__startAsyncDownloads(URLs, bypassCache, onDownloadsComplete) {
        let downloadQueue = new misc.AsyncTaskQueue(onDownloadsComplete);
        
        for each (let URL in URLs) {
            try {
                downloadQueue.addTask(new netutils.DownloadTask(URL, this._makeTempFile(URL), undefined, bypassCache));
            }
            catch (e) {
                this._logger.warn("Could not add download task for URL " + URL + ". " + e);
            }
        }
        
        downloadQueue.startTasks();
    },
    
    
    _check4NewManifests: function Updater__check4NewManifests(downloadQueue) {
        this._logger.debug("Checking manifests");
        
        let pacMan = this._application.packageManager;
        let blackList = this._application.packagesBlackList;
        let newManifests = [];
        
        for each (let mfsDlTask in downloadQueue.finishedTasks) {
            let manifestURL = mfsDlTask.uri.spec;
            
            if (mfsDlTask.statusCode === Cr.NS_OK) {
                try {
                    let packageManifest = new this._application.BarPlatform
                            .PackageManifest(manifestURL, mfsDlTask.outputFile, this._application.isTrustedPackageURL(manifestURL));
                    let [newPkgInfo, ] = this._application.selectBestPackage(packageManifest);
                    if (!newPkgInfo)
                        continue;
                    
                    if (pacMan.isPackageInstalled(manifestURL)) {
                        let currPkgInfo = pacMan.getPackageInfo(manifestURL);
                        if (this._comparator.compare(newPkgInfo.version, currPkgInfo.version) <= 0)
                            continue;
                    }
                    
                    let packageBanReason = blackList.banReasonForPackage(newPkgInfo);
                    if (packageBanReason) {
                        this._logger.warn( strutils.formatString("New package '%1' is banned (%2). Skipping it.",
                                                                 [manifestURL, packageBanReason]) );
                        continue;
                    }
                    newManifests.push({manifest: packageManifest, suitablePkgInfo: newPkgInfo});
                }
                catch(e) {
                    this._logger.warn("Could not check new manifest from URL [" + manifestURL + "] \n" + strutils.formatError(e));
                }
                finally {
                    fileutils.removeFileSafe(mfsDlTask.outputFile);
                }
            }
            else {
                this._logger.warn(strutils.formatString(
                    this._consts.ERR_MFS_DL,
                    [manifestURL, mfsDlTask.statusCode, mfsDlTask.error]));
                fileutils.removeFileSafe(mfsDlTask.outputFile);
            }
        }
        
        return newManifests;
    },
    
    
    _preinstallPackage: function Updater__preinstallPackage(packageInfo, archiveFile) {
        let preinstDir = archiveFile.clone();
        preinstDir.createUnique(Ci.nsIFile.DIRECTORY_TYPE, parseInt("0755", 8));
        fileutils.extractZipArchive(archiveFile, preinstDir);
        let preinstPkg = new this._application.BarPlatform.FilePackage(preinstDir, packageInfo.id, packageInfo.permissions);
        return preinstPkg;
    },
    
    
    _checkComponents: function Updater__checkComponents(expectedWidgets, expectedPlugins, newPackages) {
        const BarPlatform = this._application.BarPlatform;
        const widgetLibrary = this._application.widgetLibrary;
        const pacMan = this._application.packageManager;
        
        let removedWIDs = [],
            removedPIDs = [],
            missingWIDs = [],
            missingPIDs = [];
        
        this._logger.debug("Expecting widgets: " + misc.mapKeysToArray(expectedWidgets) +
                           ". Expecting plugins: " + misc.mapKeysToArray(expectedPlugins));
        for each (let widgetID in widgetLibrary.getWidgetIDs()) {
            try {
                let [packageID, widgetName] = BarPlatform.parseComponentID(widgetID);
                if (!(packageID in newPackages)) continue;
                let unit = newPackages[packageID].tempPackage.getUnit(widgetName);
                unit.checkSecurity();
                if (unit.componentInfo.type != "widget")
                    removedWIDs.push(widgetID);
            }
            catch (e) {
                this._logger.warn(strutils.formatString(this._consts.ERR_WUNIT_FAIL, [widgetID, e]));
                removedWIDs.push(widgetID);
            }
        }
        for each (let pluginID in widgetLibrary.getPluginIDs()) {
            try {
                let [packageID, pluginName] = BarPlatform.parseComponentID(pluginID);
                if (!(packageID in newPackages)) continue;
                let unit = newPackages[packageID].tempPackage.getUnit(pluginName);
                unit.checkSecurity();
                if (unit.componentInfo.type != "plugin")
                    removedPIDs.push(pluginID);
            }
            catch (e) {
                this._logger.warn(strutils.formatString(this._consts.ERR_PUNIT_FAIL, [pluginID, e]));
                removedPIDs.push(pluginID);
            }
        }
        for (let widgetID in expectedWidgets) {
            try {
                let [packageID, widgetName] = BarPlatform.parseComponentID(widgetID);
                let isNewPackage = (packageID in newPackages);
                let checkedPackage = isNewPackage ? newPackages[packageID].tempPackage : pacMan.getPackage(packageID);
                let unit = checkedPackage.getUnit(widgetName);
                if (isNewPackage)
                    unit.checkSecurity();
                if (unit.componentInfo.type != "widget")
                    missingWIDs.push(widgetID);
            }
            catch (e) {
                this._logger.warn(strutils.formatString(this._consts.ERR_WUNIT_FAIL, [widgetID, e]));
                missingWIDs.push(widgetID);
            }
        }
        for (let pluginID in expectedPlugins) {
            try {
                let [packageID, pluginName] = BarPlatform.parseComponentID(pluginID);
                let isNewPackage = (packageID in newPackages);
                let checkedPackage = isNewPackage ? newPackages[packageID].tempPackage : pacMan.getPackage(packageID);
                let unit = checkedPackage.getUnit(pluginName);
                if (isNewPackage)
                    unit.checkSecurity();
                if (unit.componentInfo.type != "plugin")
                    missingPIDs.push(pluginID);
            }
            catch (e) {
                this._logger.warn(strutils.formatString(this._consts.ERR_PUNIT_FAIL, [pluginID, e]));
                missingPIDs.push(pluginID);
            }
        }
        
        return [removedWIDs, removedPIDs, missingWIDs, missingPIDs];
    },
    
    
    _installUpdatePackage: function Updater__installUpdatePackage(preinstDir, packageInfo) {
        let widgetLibrary = this._application.widgetLibrary;
        let packageID = packageInfo.id;
        let prevPluginsState = widgetLibrary.getCurrentPluginsState(packageID);
        this._application.switchWidgets(packageID, false);
        
        try {
            widgetLibrary.flushWidgets(packageID);
            widgetLibrary.cleanPackageParserCache(packageID);
            widgetLibrary.flushPlugins(packageID);
            
            this._application.packageManager.installPackage(preinstDir, packageInfo);
        }
        finally {
            widgetLibrary.setPluginsState(prevPluginsState, true);
            this._application.switchWidgets(packageID, true);
        }
    },
    
    _storePreset: function  Updater__storePreset(preset) {
        let presetFile = this._application.directories.presetsDir;
        presetFile.append(encodeURIComponent(preset.address));
        preset.saveToFile(presetFile);
    },
    
    _compareBlacklistEffect: function Updater__compareBlacklistEffect(oldBlacklist, newBlacklist) {
        let pacMan = this._application.packageManager;
        let changedPackages = [];
        for each (let packageID in pacMan.packageIDs) {
            let packageInfo = pacMan.getPackageInfo(packageID);
            let lastReason = oldBlacklist.banReasonForPackage(packageInfo);
            let newReason = newBlacklist.banReasonForPackage(packageInfo);
            if (!lastReason !== !newReason)
                changedPackages.push(packageInfo);
        }
        
        return changedPackages;
    },
    
    _startUpdates: function Updater__startUpdates(updateController) {
        this._logger.debug("Updates check initiated");
        
        try {
            this._updateBlackList(
                sysutils.bindFunction(this._updateComponents, this, [updateController]) );
        }
        finally {
            this._timer.cancel();
            this._timer.initWithCallback(this, this._updateInterval, this._timer.TYPE_ONE_SHOT);
            this._lastUpdateTimestamp = Date.now();
        }
    },
    
    _updateComponents: function Updater__updateComponents(updateController) {
        let updProcess = new UpdateProcess(updateController);
        updateController.start(updProcess);
    },
    
    _updateBlackList: function Updater__updateBlackList(onDone) {
        this._downloadBlackList(onDone);
    },
    
    _blackListURL: "http://export.yandex.ru/bar/blacklist.xml",
    
    _downloadBlackList: function Updater__checkBlackListUpdates(onDone) {
        this._logger.debug("Downloading package black list from " + this._blackListURL);
        
        let downloadQueue = new misc.AsyncTaskQueue(
            sysutils.bindFunction(this._onBlackListDownloaded, this, [onDone]) );
        downloadQueue.addTask(new netutils.DownloadTask(this._blackListURL, this._makeTempFile(this._blackListURL),
                                                        undefined, true, {"If-None-Match": this._blackListETag}) );
        downloadQueue.startTasks();
    },
    
    _onBlackListDownloaded: function Updater__onBlackListDownloaded(onDone, dlQueue) {
        let packageInfos;
        try {
            let blackListDlTask = dlQueue.finishedTasks[0];
            if (blackListDlTask.statusCode == Cr.NS_OK) {
                let responseStatus = blackListDlTask.httpStatus;
                if (responseStatus >= 200 && responseStatus < 300) {
                    let blDoc = this._application.core.Lib.fileutils.xmlDocFromFile(blackListDlTask.outputFile);
                    let blackList = new this._application.BarPlatform.BlackList.fromDoc(blDoc);
                    let prevBlackList = this._application.packagesBlackList;
                    
                    this._application.packagesBlackList = blackList;
                    this._blackListETag = blackListDlTask.findHttpResponseHeader("ETag") || "";
                    
                    packageInfos = this._compareBlacklistEffect(prevBlackList, blackList);
                }
            }
            else
                this._logger.warn("Black list was not downloaded. Status code: " + blackListDlTask.statusCode);
        }
        catch (e) {
            this._logger.error("Could not update black list from [" + this._blackListURL + "] \n"+ strutils.formatError(e));
        }
        
        if (packageInfos && packageInfos.length) {
            let packageIDs = packageInfos.map(function (info) info.id);
            this._application.forEachWindow(function(overlay) {
                try {
                    overlay.deleteObsoleteWidgetItems();
                    overlay.updatePalette(packageIDs);
                }
                catch (e) {
                    this._logger.error("Could not update window or palette. " + strutils.formatError(e));
                    this._logger.debug(e.stack);
                }
            }, this);
            for each (let packageID in packageIDs) {
                this._application.widgetLibrary.flushWidgets(packageID);
            }
            
            this._showBanPopupNotification(packageInfos, onDone);
        }
        else {
            if (onDone)
                onDone();
        }
    },
    
    _showBanPopupNotification: function Updater__showBanPopupNotification(packages, onDone) {
        this._logger.debug("Bans updated, ask user");
        
        misc.openWindow({
            url: "chrome://" + this._application.name + "/content/custombar/dialogs/package-management/ban/ban-notification.xul",
            features: "__popup__",
            name: "ban-notification",
            application: this._application,
            then: sysutils.bindFunction(this._onBanPopupClosed, this, [packages, onDone])
        });
    },
    _onBanPopupClosed: function Updater__onBanPopupClosed(packages, onDone, showBans) {
        if (!showBans) {
            if (onDone)
                onDone();
            return;
        }
        
        this._application.core.Lib.misc.openWindow({
            url: "chrome://" +
                 this._application.name +
                 "/content/custombar/dialogs/package-management/ban/ban.xul",
            features: "__popup__",
            name: "package-management-ban",
            packages: packages,
            application: this._application,
            then: onDone
        });
    },
    
    
    
    riseSkippedUpdatesThrottling: function Updater_riseSkippedUpdatesThrottling(count) {
        for (let i= 0; i < count; ++i) {
            if (!this._updatesThrottled) break;
            this._updatesThrottlingTick();
        }
    },
    
    resetUpdatesThrottling: function Updater_resetUpdatesThrottling() {
        this._updatesThrottling = 2;
    },
    
    get _updatesThrottled() {
        return this._updatesThrottling & (this._updatesThrottling - 1);
    },
    
    _updatesThrottlingTick: function Updater__updatesThrottlingTick() {
        let current = this._updatesThrottling + 1;
        let limit = this._updatesThrottlingLimit;
        if (current > limit * 2) {
            current-= limit;
        }
        this._updatesThrottling = current;
    },
    
    get _updatesThrottling() {
        return this._application.preferences.get(this._updatesThrottlingPrefName, 2);
    },
    
    set _updatesThrottling(val) {
        this._application.preferences.set(this._updatesThrottlingPrefName, val);
    },
    
    get _updatesThrottlingLimit() {
        return this._application.preferences.get("barplatform.components.updates.throttling.limit", 32);
    },
    
    get _updatesThrottlingPrefName() {
        return "barplatform.components.updates.throttling";
    }
};

const UpdateProcess = function UpdateProcess(updateController) {
    this._updateController = updateController;
    
    this._application = updater._application;
    this._logger = this._application.getLogger("CompUpdater");
};

UpdateProcess.prototype = {
    constructor: UpdateProcess,
    
    checkPresets: function UpdateProcess_checkPresets(defaultOnly) {
        this._downloadPresets(defaultOnly);
    },
    
    startSilentDefaultPresetUpdate: function UpdateProcess_startSilentDefaultPresetUpdate() {
        if (this._defPresetPair) {
            this._startSilentUpdate();
        }
        else {
            this._logger.debug("No new version of default preset available");
            this._updateController.onDefaultPresetUpdateFinish();
        }
    },
    
    
    startUserComponentsUpdate: function UpdateProcess_startUserComponentsUpdate(newPresetPairs) {
        this._startUserCompsUpdate(newPresetPairs);
    },
    
    _updateController: null,
    _presetsChecked: null,
    _defPresetPair: null,
    _newWIDsSet: null,
    _newPIDsSet: null,
    _defPresetPkgIDsSet: null,
    _newPkgsInfo: null,
    _updatedPackageIDs: null,
    _newPackages: null,
    
    _downloadPresets: function UpdateProcess__downloadPresets(defaultOnly) {
        this._logger.debug("Downloading presets");
        
        this._presetsChecked = updater._getInstalledPresets(defaultOnly);
        
        let downloadQueue = new misc.AsyncTaskQueue(
            sysutils.bindFunction(this._onPresetsDownloaded, this));
        
        let mark = {};
        mark[this._application.core.presetRequestMarkName] = true;
        for (let presetURL in this._presetsChecked) {
            try {
                downloadQueue.addTask(new netutils.DownloadTask(presetURL, updater._makeTempFile(presetURL), mark, true));
            }
            catch (e) {
                this._logger.warn("Could not add download task for preset with URL " + presetURL + ". " + e);
            }
        }
        
        downloadQueue.startTasks();
    },
    
    _onPresetsDownloaded: function UpdateProcess__onPresetsDownloaded(downloadQueue) {
        this._logger.debug("Checking presets");
        
        let updPresets = [];
        for each (let presetDlTask in downloadQueue.finishedTasks) {
            let presetURL = presetDlTask.uri.spec;
            
            if (presetDlTask.statusCode === Cr.NS_OK) {
                try {
                    let newPreset = new this._application.BarPlatform.Preset(presetDlTask.outputFile, presetURL);
                    let oldPreset = this._presetsChecked[presetURL];
                    
                    if (updater._comparator.compare(newPreset.version, oldPreset.version) > 0) {
                        if (!this._defPresetPair
                                && newPreset.updateMode != "default"
                                && presetURL == this._application.defaultPresetURL)
                            this._defPresetPair = {newPreset: newPreset, oldPreset: oldPreset};
                        else
                            updPresets.push({newPreset: newPreset, oldPreset: oldPreset});
                    }
                }
                catch(e) {
                    this._logger.warn("Could not check new preset from " + presetURL + ". " + strutils.formatError(e));
                }
                finally {
                    fileutils.removeFileSafe(presetDlTask.outputFile);
                }
            }
            else {
                this._logger.warn(strutils.formatString(updater._consts.ERR_PRESET_DL,
                    [presetURL, presetDlTask.statusCode, strutils.formatError(presetDlTask.error)]));
                fileutils.removeFileSafe(presetDlTask.outputFile);
            }
        }
        
        try {
            this._updateController.onPresetsDownloaded(this._defPresetPair, updPresets);
        }
        catch (e) {
            this._logger.error("Update failed. " + strutils.formatError(e));
            this._logger.debug(e.stack);
        }
    },
    
    _startSilentUpdate: function UpdateProcess__startSilentUpdate() {
        this._logger.debug("Silent update started");
        
        this._defPresetPkgIDsSet = {};
        this._newWIDsSet = {};
        this._newPIDsSet = {};
        
        updater._gatherPresetInfo(this._defPresetPair.newPreset, this._defPresetPkgIDsSet,
                                  this._newWIDsSet, this._newPIDsSet);
        
        let URLsList = misc.mapKeysToArray(this._defPresetPkgIDsSet);
        this._logger.debug("Downloading package manifests: " + URLsList);
        updater._startAsyncDownloads(URLsList, true, sysutils.bindFunction(this._silentDownloadPkgs, this) );
    },
    _silentDownloadPkgs: function UpdateProcess__silentDownloadPkgs(mfsQueue) {
        let newMfsInf = updater._check4NewManifests(mfsQueue);
        let newPkgsInfo = this._newPkgsInfo = [];
        let newPkgsURLs = [];
        for each (let pkgElem in newMfsInf) {
            let pkgInfo = pkgElem.suitablePkgInfo;
            newPkgsInfo.push(pkgInfo);
            newPkgsURLs.push(pkgInfo.fileURL);
        }
        
        if (newPkgsURLs.length > 0) {
            this._logger.debug("Downloading package archives: " + newPkgsURLs);
            updater._startAsyncDownloads(newPkgsURLs, true, sysutils.bindFunction(this._onPackagesDownloaded, this) );
        }
        else {
            this._logger.debug("No new packages are available right now.");
            this._onPackagesDownloaded();
        }
    },
    
    _onPackagesDownloaded: function UpdateProcess__onPackagesDownloaded(pkgsQueue) {
        let newPkgsInfo = this._newPkgsInfo;
        function findPkgInfo(archiveURL) {
            for each (let packageInfo in newPkgsInfo) {
                if (packageInfo.fileURL == archiveURL)
                    return packageInfo;
            }
            return null;
        }
        
        const pacMan = this._application.packageManager;
        
        let updatedPackageIDs = this._updatedPackageIDs = [];
        this._newPackages = {};
        
        if (pkgsQueue) {
            for each (let pkgDlTask in pkgsQueue.finishedTasks) {
                try {
                    let archiveURL = pkgDlTask.uri.spec;
                    
                    if (pkgDlTask.statusCode !== Cr.NS_OK) {
                        this._logger.warn(strutils.formatString(
                            updater._consts.ERR_PKG_DL,
                            [archiveURL, pkgDlTask.statusCode, pkgDlTask.error]));
                        continue;
                    }
                    
                    let packageInfo = findPkgInfo(archiveURL);
                    if (!packageInfo) {
                        this._logger.warn("Strange thing happened. Could not find PackageInfo for \"" + archiveURL + "\"");
                        continue;
                    }
                    
                    try {
                        let tempPackage = updater._preinstallPackage(packageInfo, pkgDlTask.outputFile);
                        
                        let packageID = packageInfo.id;
                        this._newPackages[packageID] = {tempPackage: tempPackage, packageInfo: packageInfo};
                        if (pacMan.isPackageInstalled(packageID))
                            updatedPackageIDs.push(packageID);
                    }
                    catch (e) {
                        this._logger.error("Could not preinstall package from " + archiveURL + ". " + strutils.formatError(e));
                    }
                }
                finally {
                    fileutils.removeFileSafe(pkgDlTask.outputFile);
                }
            }
        }
        
        this._silentUpdate();
    },
    
    _silentUpdate: function UpdateProcess__silentUpdate() {
        let widgetLibrary = this._application.widgetLibrary;
        let widgetsChanged = 0,
            pluginsChanged = 0;
        try {
            let [removedWIDs, removedPIDs,
                 missingWIDs, missingPIDs] = updater._checkComponents(this._newWIDsSet, this._newPIDsSet, this._newPackages);
            
            if (missingWIDs.length > 0)
                throw new Error("Missing widgets: " + missingWIDs);
            if (missingPIDs.length > 0)
                throw new Error("Missing plugins: " + missingPIDs);
            
            for each (let newPkgData in this._newPackages)
                updater._installUpdatePackage(newPkgData.tempPackage.rootDirectory, newPkgData.packageInfo);
            widgetsChanged += widgetLibrary.forgetWidgets(removedWIDs, true);
            pluginsChanged += widgetLibrary.forgetPlugins(removedPIDs, true);
            let staticWidgets = this._application.overlayProvider.staticWidgets;
            for (let widgetID in this._newWIDsSet) {
                if (widgetID in staticWidgets) continue;
                widgetsChanged += widgetLibrary.registerWidgets(widgetID, true, true);
            }
            for (let pluginID in this._newPIDsSet)
                pluginsChanged += widgetLibrary.registerPlugins(pluginID, true, true);
            
            widgetLibrary.persist(widgetsChanged > 0, pluginsChanged > 0);
            this._application.forEachWindow(
                function UpdateProcess_updBrowserWnds(controller) {
                    controller.updatePalette(this._updatedPackageIDs);
                    controller.deleteObsoleteWidgetItems();
                },
                this, true
            );
            
            this._defaultPresetUpdate(this._defPresetPair.newPreset, this._newWIDsSet, this._newPIDsSet);
        }
        catch (e) {
            this._logger.error("Silent update failed. " + strutils.formatError(e));
            this._logger.debug(e.stack);
        }
        finally {
            for each (let newPkgData in this._newPackages) {
                let rootDir = newPkgData.tempPackage.rootDirectory;
                newPkgData.tempPackage.finalize();
                fileutils.removeFileSafe(rootDir);
            }
        }
        
        this._updateController.onDefaultPresetUpdateFinish(this._defPresetPkgIDsSet);
    },
    _defaultPresetUpdate: function UpdateProcess__defaultPresetUpdate(newPreset, newWIDsSet, newPIDsSet) {
        const updateApplyModeMap = {"reset" : "hard", "soft-reset": "soft", "reorder": "order", "silent": "addleft"};
        const widgetLibrary = this._application.widgetLibrary;
        const pacMan = this._application.packageManager;
        const BarPlatform = this._application.BarPlatform;
        
        let resetMode = (newPreset.updateMode == "reset"),
            softResetMode = (newPreset.updateMode == "soft-reset"),
            reorderMode = (newPreset.updateMode == "reorder");
        
        let prevDefPresetComps = this._application.defaultPreset.componentIDs;
        let newDefPresetWidgets = newPreset.widgetIDs;
        let staticWidgets = this._application.overlayProvider.staticWidgets;
        for (let widgetID in newDefPresetWidgets) {
            if (!(widgetID in prevDefPresetComps) && (widgetID in staticWidgets))
                newWIDsSet[widgetID] = 1;
        }
        this._application.defaultPreset = newPreset;
        
        if (resetMode)
            BarPlatform.eraseSettings();
        let applyMode = updateApplyModeMap[newPreset.updateMode] || "addleft";
        this._application.forEachWindow (
            function UpdateProcess_applyPresetToWindow(controller) {
                controller.reloadStaticWidgets();
                controller.applyPreset(newPreset, applyMode, newWIDsSet, true);
                controller.updateToolbarDefaultSet();
            }
        );
        
        if (!resetMode) {
            for each (let pluginEntry in newPreset.pluginEntries) {
                let isNewPlugin = (pluginEntry.componentID in newPIDsSet);
                let plugin = this._application.widgetLibrary.getPlugin(pluginEntry.componentID);
                
                plugin.applySettings(isNewPlugin ? pluginEntry.settings : pluginEntry.forcedSettings, true);
                if (isNewPlugin || reorderMode)
                    plugin.enabled = (pluginEntry.enabled == pluginEntry.ENABLED_YES);
            }
        }
        
        if (resetMode || softResetMode) {
            widgetLibrary.setDefaultPluginsState(softResetMode);
            let newDefPresetComps = newPreset.componentIDs;
            let affectedPackages = {};
            for (let widgetID in prevDefPresetComps) {
                try {
                    if (!(widgetID in newDefPresetComps) && !(widgetID in staticWidgets)) {
                        widgetLibrary.forgetComponents([widgetID]);
                        let [packageID, ] = BarPlatform.parseComponentID(widgetID);
                        affectedPackages[packageID] = 1;
                    }
                }
                catch (e) {
                    this._logger.error("Could not uninstall component. " + strutils.formatError(e));
                }
            }
            for (let packageID in affectedPackages) {
                if (widgetLibrary.getWidgetIDs(packageID).length < 1) {
                    pacMan.uninstallPackage(packageID);
                    widgetLibrary.cleanPackageParserCache(packageID);
                }
            }
        }
        
        updater._storePreset(newPreset);
        
        updater._obsService.notifyObservers(null, this._application.core.eventTopics.EVT_AFTER_DEFPRESET_UPD, null);
    },
    
    _startUserCompsUpdate: function UpdateProcess__startUserCompsUpdate(updPresets) {
        this._logger.debug("Explicit update started. We have " + updPresets.length + " new presets");
        let packageIDsSet = {};
        
        let checkedIDsSet = this._defPresetPkgIDsSet;
        let newPresets = [];
        for each (let presetPair in updPresets) {
            let preset = presetPair.newPreset;
            newPresets.push(preset);
            for (let packageID in preset.packageIDs) {
                if (!checkedIDsSet || !(packageID in checkedIDsSet))
                    packageIDsSet[packageID] = 1;
            }
        }
        for each (let [, packageID] in Iterator(this._application.packageManager.packageIDs)) {
            if (!checkedIDsSet || !(packageID in checkedIDsSet))
                packageIDsSet[packageID] = 1;
        }
        
        let URLsList = misc.mapKeysToArray(packageIDsSet);
        this._logger.debug("Need to check " + URLsList.length + " packages");
        if (URLsList.length > 0)
            this._logger.debug("Downloading package manifests: " + URLsList);
        updater._startAsyncDownloads(URLsList, true,
                                     sysutils.bindFunction(this._explicitUpdate, this, [newPresets]));
    },
    
    _explicitUpdate: function UpdateProcess__explicitUpdate(newPresets, mfsQueue) {
        let newManifests = updater._check4NewManifests(mfsQueue).map(function(infoEl) infoEl.manifest);
        this._updateController.onExplicitUpdateDataReady(newPresets, newManifests);
    }
};
