"use strict";

const {
    classes: Cc,
    interfaces: Ci,
    results: Cr,
    utils: Cu
} = Components;
const EXTENSION_DIR = __LOCATION__.parent.parent;
const EXTENSION_PATH = Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService).newFileURI(EXTENSION_DIR).spec;
Cc["@mozilla.org/moz/jssubscript-loader;1"].getService(Ci.mozIJSSubScriptLoader).loadSubScript(EXTENSION_PATH + "config.js");

Cu.import("resource://gre/modules/XPCOMUtils.jsm");
function CustomBarCore() {
    this._loadModules();
    
    this._observerService.addObserver(this, "profile-after-change", false);
    this._observerService.addObserver(this, "profile-before-change", false);
    this._observerService.addObserver(this, "quit-application", false);
};

CustomBarCore.prototype = {
    
    Lib: {},
    
    get appName() {
        return this._appName;
    },
    get version() {
        return this._version;
    },
    
    get buidDate() {
        return new Date(this._buildTimeStamp);
    },
    
    get buidRevision() {
        return BUILD_REVISION;
    },
    
    get xbWidgetsPrefsPath() {
        return XB_WIDGETS_PREFS_PATH;
    },
    
    get nativesPrefsPath() {
        return NATIVES_PREFS_PATH;
    },
    
    get staticPrefsPath() {
        return STATIC_PREFS_PATH;
    },
    
    get application() {
        return this._appObj;
    },
    
    get rootDir() {
        if (!this._appRoot) {
            this._appRoot = Cc["@mozilla.org/file/directory_service;1"].getService(Ci.nsIProperties).get("ProfD", Ci.nsIFile);
            this._appRoot.append(this.appName + "-xb");
            if (!this._appRoot.exists())
                this._appRoot.create(Ci.nsIFile.DIRECTORY_TYPE, parseInt("0755", 8));
        }
        return this._appRoot.clone();
    },
    
    get xbProtocol() {
        return Cc[PROTOCOL_XB_CONTRACT_ID].getService(Ci.nsIProtocolHandler).QueryInterface(Ci.nsIXBProtocolHandler);
    },
    
    get presetRequestMarkName() {
        return this._presetRequestMarkName;
    },
    
    get extensionPathFile() {
        return EXTENSION_DIR.clone();
    },
    get wrappedJSObject() {
        return this;
    },
    
    get eventTopics() {
        return this._globalEvents;
    },
    
    get logging() {
        return this._logging;
    },
    
    set logging(newVal) {
        if (this._logging == !!newVal)
            return;
        
        this._logging = !this._logging;
        if (this._logging)
            this._startLogging();
        else
            this._stopLogging();
    },
    observe: function CustomBarCore_observe(subject, topic, data) {
        switch (topic) {
            case "http-on-modify-request":
                try {
                    this._checkPresetRequest(subject.QueryInterface(Ci.nsIHttpChannel))
                }
                catch (e) {
                    this._logger.error("Error observing 'http-on-modify-request'. " + this._formatError(e));
                    this._logger.debug(e.stack);
                }
                break;
            
            case "profile-after-change":
                this._startLogging();
                this._initApp();
                this._observerService.addObserver(this, "http-on-modify-request", false);
                break;
                
            case "profile-before-change":
                this._shutdown();
                this._stopLogging();
                break;
            
            case "nsPref:changed":
                this._updateLoggers(data == this._dumpLogLevelPrefName,
                                    data == this._consoleLogLevelPrefName,
                                    data == this._fileLogLevelPrefName);
                break;
        }
    },
    get oldCore() {
        try {
            let oldCore = ("nsIYaSearch" in Ci) && Cc["@yandex.ru/" + APP_NAME + ";1"].getService(Ci.nsIYaSearch).wrappedJSObject;
            this.__defineGetter__("oldCore", function() oldCore);
            return this.oldCore;
        } catch (e) {}
        
        return null;
    },
    _version: PLATFORM_VERSION,
    _buildTimeStamp: Date.parse(BUILD_DATE),
    _appName: APP_NAME,
    _modulesPath: "resource://" + APP_NAME + "-mod/",
    _appResPath: "resource://" + APP_NAME + "-app/",
    _appChromePath: "chrome://" + APP_NAME + "/",
    
    _appRoot: null,
    _appObj: null,
    
    _dumpLogLevelPrefName: APP_NAME + ".xbcore.logging.stdout.level",
    _consoleLogLevelPrefName: APP_NAME + ".xbcore.logging.console.level",
    _fileLogLevelPrefName: APP_NAME + ".xbcore.logging.file.level",
    _dumpAppender: null,
    _consoleAppender: null,
    _fileAppender: null,
    _logging: true,
    
    
    _globalEvents: {
        EVT_PLUGIN_BEFORE_ENABLED: APP_NAME + "-platform-plugin-before_enabled",
        EVT_PLUGIN_ENABLED: APP_NAME + "-platform-plugin-enabled",
        EVT_PLUGIN_BEFORE_DISABLED: APP_NAME + "-platform-plugin-before_disabled",
        EVT_PLUGIN_DISABLED: APP_NAME + "-platform-plugin-disabled",
        EVT_AFTER_DEFPRESET_UPD: APP_NAME + "-platform-defpreset-after_update",
        EVT_BEFORE_GLOBAL_RESET: APP_NAME + "-platform-components-before_reset",
        EVT_AFTER_GLOBAL_RESET: APP_NAME + "-platform-components-after_reset"
        
    },
    _moduleFileNames: [
        "Log4moz.jsm", "Preferences.jsm", "WindowListener.jsm", "AddonManager.jsm", "Foundation.jsm",
        "JSONvsXML.jsm", "wc.jsm"
    ],
    
    _observerService: Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService),
    _presetURLPattern: /^.+\.xb\.xml$/,
    _presetRequestMarkName: APP_NAME + "-preset-download",
    
    _loadModules: function CustomBarCore__loadModules() {
        const Lib = this.Lib;
        const prePath = this._modulesPath;
        this._moduleFileNames.forEach(function loadModule(fileName) {
            Cu.import(prePath + fileName, Lib);
        });
    },
    
    _startLogging: function CustomBarCore__startLogging() {
        const Log4Moz = this.Lib.Log4Moz;
        
        let root = Log4Moz.repository.rootLogger;
        root.level = Log4Moz.Level.All;
        
        let formatter = new Log4Moz.BasicFormatter();
        this._consoleAppender = new Log4Moz.ConsoleAppender(formatter);
        root.addAppender(this._consoleAppender);
        this._dumpAppender = new Log4Moz.DumpAppender(formatter);
        root.addAppender(this._dumpAppender);
        let logFile = this.rootDir;
        logFile.append("debug.log");
        this._fileAppender = new Log4Moz.RotatingFileAppender(logFile, formatter);
        root.addAppender(this._fileAppender);
        this._logger = Log4Moz.repository.getLogger(this.appName + ".Core");
        this._logger.level = Log4Moz.Level.Debug;
        this._updateLoggers(true, true, true);
        
        this.Lib.Preferences.observe(this._dumpLogLevelPrefName, this);
        this.Lib.Preferences.observe(this._consoleLogLevelPrefName, this);
        this.Lib.Preferences.observe(this._fileLogLevelPrefName, this);
    },
    
    _stopLogging: function CustomBarCore__stopLogging() {
        this.Lib.Preferences.ignore(this._dumpLogLevelPrefName, this);
        this.Lib.Preferences.ignore(this._consoleLogLevelPrefName, this);
        this.Lib.Preferences.ignore(this._fileLogLevelPrefName, this);
        
        let root = this.Lib.Log4Moz.repository.rootLogger;
        
        root.removeAppender(this._consoleAppender);
        this._consoleAppender = null;
        
        root.removeAppender(this._dumpAppender);
        this._dumpAppender = null;
        
        this._fileAppender.closeStream();
        root.removeAppender(this._fileAppender);
        this._fileAppender = null;
    },
    
    _updateLoggers: function CustomBarCore__updateLoggers(checkDumpSetting, checkConsoleSetting, checkFileSetting) {
        const Preferences = this.Lib.Preferences;
        function readLogLevel(prefName) {
            let logLevel = parseInt( Preferences.get(prefName, NaN), 10 );
            return (isNaN(logLevel) || (logLevel < 0)) ? 100 : logLevel;
        }
        
        if (checkDumpSetting)
            this._dumpAppender.level = readLogLevel(this._dumpLogLevelPrefName);
        
        if (checkConsoleSetting)
            this._consoleAppender.level = readLogLevel(this._consoleLogLevelPrefName);
        
        if (checkFileSetting)
            this._fileAppender.level = readLogLevel(this._fileLogLevelPrefName)
    },
    
    _initApp: function CustomBarCore__initApp() {
        try {
            let appModule = {};
            Cu.import(this._appResPath + "bar.js", appModule);
            this._appObj = appModule.barApplication;
            this._appObj.init(this);
            
            this._observerService.notifyObservers(null, this.appName + "-state-changed", "custombar-initialized");
        }
        catch (e) {
            this._appObj = null;
            this._logger.fatal("Couldn't initialize application. " + this._formatError(e));
            if (e.stack)
                this._logger.debug(e.stack);
        }
    },
    
    _destroyApp: function CustomBarCore__destroyApp() {
        try {
            this._appObj.finalize();
            this._appObj = null;
        } catch(e) {
            this._logger.error(this._formatError(e));
            this._logger.debug(e.stack);
        }
    },
    
    _shutdown: function CustomBarCore__shutdown() {
        this._observerService.removeObserver(this, "http-on-modify-request");
        
        if (this._appObj)
            this._destroyApp();
    },
    
    _formatError: function CustomBarCore__formatError(e) {
        
        
        let text = e.name + ": " + e.message;
        let fileName = e.fileName || e.filename;
        if (fileName)
            text += "\nin " + fileName + "@" + e.lineNumber;
        
        return text;
    },
    
    _checkPresetRequest: function CustomBarCore__checkPresetRequest(channel) {
        if (!channel.URI.path.split(/\?|#|;/)[0].match(this._presetURLPattern))
            return;
        
        try {
            channel.QueryInterface(Ci.nsIPropertyBag2);
            if (channel.getPropertyAsBool(this._presetRequestMarkName))
                return;
        } catch (e) {}
        let presetURL = channel.URI.spec;
        channel.cancel(Cr.NS_BINDING_ABORTED);
        
        this._logger.debug("XB preset request caught. Installing: " + presetURL);
        this.application.installPreset(presetURL);
    },
    
    classDescription: "Custom Yandex bar core JS component for " + APP_NAME,
    classID: CORE_CLASS_ID,
    contractID: CORE_CONTRACT_ID,
    QueryInterface: XPCOMUtils.generateQI([Ci.nsISupports, Ci.nsIObserver]),
    _xpcom_categories: [{ category: "app-startup", service: true }]
};

if (XPCOMUtils.generateNSGetFactory)
  var NSGetFactory = XPCOMUtils.generateNSGetFactory([CustomBarCore]);
else
  var NSGetModule = XPCOMUtils.generateNSGetModule([CustomBarCore]);
