"use strict";

const netutils = {
    ioService: Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService),
    cookieService: Cc["@mozilla.org/cookieService;1"].getService().QueryInterface(Ci.nsICookieService),
    
    newURI: function netutils_newURI(spec, origCharset, baseURI) {
        return this.ioService.newURI(spec, origCharset, baseURI);
    },
    
    resolveRelativeURL: function netutils_resolveRelativeURL(spec, baseURI) {
        var uri = this.ioService.newURI(spec, null, baseURI);
        return uri.spec;
    },
    
    
    findCookieValue: function netutils_findCookieValue(URLorURI, cookieName, incHttpOnly ) {
        let cookieURI = (URLorURI instanceof Ci.nsIURI) ? URLorURI : this.newURI(URLorURI, null, null);
        
        let cookieString = incHttpOnly ?
            this.cookieService.getCookieStringFromHttp(cookieURI, null, null) :
            this.cookieService.getCookieString(cookieURI, null);
        
        if (!cookieString)
            return undefined;
        
        let matchString = "(?:^|;)\\s*" + strutils.escapeRE(cookieName) + "=([^;]*)";
        let cookieMatch = cookieString.match(matchString);
        
        return cookieMatch ? decodeURIComponent(cookieMatch[1].trim()) : undefined;
    }
};

netutils.DownloadTask = misc.AsyncTask.extend({
    
    constructor: function DownloadTask(urlString, outputFile, channelProperties, bypassCache, httpHeaders) {
        try {
            this._uri = netutils.ioService.newURI(urlString, null, null);
        }
        catch (e) {
            throw (this._error = new Error(e.message + " " + urlString));
        }
        if (!(outputFile instanceof Ci.nsIFile))
            throw this._error = new CustomErrors.EArgType("outputFile", "nsIFile", outputFile);
        if ((httpHeaders !== undefined) && !sysutils.isObject(httpHeaders))
            throw this._error = new CustomErrors.EArgType("httpHeaders", "Object", httpHeaders);
        
        this._outputFile = outputFile;
        this._channelProps = channelProperties;
        this._bypassCache = !!bypassCache;
        this._httpHeaders = httpHeaders;
    },
    
    start: function DownloadTask_start(owner) {
        try {
            this.base(owner);
            
            this._outputStream = Cc["@mozilla.org/network/file-output-stream;1"].createInstance(Ci.nsIFileOutputStream);
            let modeFlags = fileutils.MODE_WRONLY | fileutils.MODE_CREATE | fileutils.MODE_TRUNCATE;
            this._outputStream.init(this._outputFile, modeFlags, fileutils.PERMS_FILE, 0);
            
            this._channel = netutils.ioService.newChannelFromURI(this._uri);
            if (this._channelProps)
                this._writeProps(this._channelProps, this._channel);
            if (this._bypassCache)
                this._channel.loadFlags |= Ci.nsIRequest.LOAD_BYPASS_CACHE;
            if (this._httpHeaders)
                this._addHttpHeaders(this._httpHeaders, this._channel);
            
            this._channel.asyncOpen(this, null);
        }
        catch (e) {
            this._error = e;
            throw e;
        }
    },
    
    abort: function DownloadTask_abort(reason) {
        this._request.cancel(Components.results.NS_BINDING_ABORTED);
        this.base(reason);
    },
    get uri() {
        return this._uri;
    },
    
    get percentDone() {
        if (!this._total)
            return NaN;
        return (this._progress == this._total) ? 100 : Math.floor(100 * (this._progress / this._total));
    },
    
    get statusCode() {
        return this._statusCode;
    },
    
    get outputFile() {
        return this._outputFile;
    },
    
    get httpStatus() {
        let httpChannel = this._channel.QueryInterface(Ci.nsIHttpChannel);
        return httpChannel.responseStatus;
    },
    
    findHttpResponseHeader: function DonwloadTask_findHttpResponseHeader(headerName) {
        let httpChannel = this._channel.QueryInterface(Ci.nsIHttpChannel);
        try {
            return httpChannel.getResponseHeader(headerName);
        }
        catch (e) {
            return undefined;
        }
    },
    QueryInterface: function DownloadTask_QueryInterface(interfaceId) {
        if (interfaceId.equals(Ci.nsISupports) ||
            interfaceId.equals(Ci.nsIStreamListener) ||
            interfaceId.equals(Ci.nsIRequestObserver) )
            return this;
        throw Cr.NS_ERROR_NO_INTERFACE;
    },
    onDataAvailable: function DownloadTask_onDataAvailable(request, context, inputStream, offset, count) {
        this._binInputStream.setInputStream(inputStream);
        this._outputStream.write(this._binInputStream.readBytes(count), count);
        this._step(count);
    },
    onStartRequest: function DownloadTask_onStartRequest(request, context) {
        if (request.status != Cr.NS_OK) {
            this._statusCode = request.status;
            return;
        }
        
        this._channel = request.QueryInterface(Ci.nsIChannel);
        this._request = request;
        this._total = this._channel.contentLength;
        this._binInputStream = Cc["@mozilla.org/binaryinputstream;1"].createInstance(Ci.nsIBinaryInputStream);
    },
    
    onStopRequest: function DownloadTask_onStopRequest(request, context, statusCode) {
        try {
            if (statusCode === Cr.NS_OK)
                this._total = this._progress;
            this._statusCode = statusCode;
            
            if (this._total && this._binInputStream)
                this._binInputStream.close();
            
            this._outputStream.close();
        }
        catch (e) {
            this._error = e;
        }
        finally {
            this._finish();
        }
    },
    
    _binInputStream: null,
    _outputStream: null,
    _outputFile: null,
    _bypassCache: false,
    _httpHeaders: null,
    _channel: null,
    _request: null,
    _total: 0,
    _statusCode: undefined,
    
    _writeProps: function DownloadTask__writeProps(propsMap, toChannel) {
        let propBag = toChannel.QueryInterface(Ci.nsIWritablePropertyBag2);
        for each(let [propName, propValue] in Iterator(propsMap)) {
            if (sysutils.isBool(propValue))
                propBag.setPropertyAsBool(propName, propValue);
            else if (sysutils.isNumber(propValue))
                propBag.setPropertyAsInt64(propName, propValue);
            else if (sysutils.isString(propValue))
                propBag.setPropertyAsAUTF8String(propName, propValue);
            else
                throw new TypeError("Can't add property of type " + (typeof propValue));
        }
    },
    
    _addHttpHeaders: function DownloadTask__addHttpHeaders(headersMap, toChannel) {
        let httpChannel = toChannel.QueryInterface(Ci.nsIHttpChannel);
        for each (let [headerName, headerValue] in Iterator(headersMap))
            httpChannel.setRequestHeader(headerName, headerValue, true);
    }
});